## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""Devices (not receivers) known to Solaar.

Solaar can handle many recent devices without having any entry here.
An entry should only be added to fix problems, such as
- the device's device ID or WPID falls outside the range that Solaar searches
- the device uses a USB interface other than 2
- the name or codename should be different from what the device reports
"""

from .hidpp10_constants import DEVICE_KIND
from .hidpp10_constants import Registers as Reg


class _DeviceDescriptor:
    def __init__(
        self,
        name=None,
        kind=None,
        wpid=None,
        codename=None,
        protocol=None,
        registers=None,
        usbid=None,
        interface=None,
        btid=None,
    ):
        self.name = name
        self.kind = kind
        self.wpid = wpid
        self.codename = codename
        self.protocol = protocol
        self.registers = registers
        self.usbid = usbid
        self.interface = interface
        self.btid = btid
        self.settings = None


DEVICES_WPID = {}
DEVICES = {}


def _D(
    name,
    codename=None,
    kind=None,
    wpid=None,
    protocol=None,
    registers=None,
    settings=None,
    usbid=None,
    interface=None,
    btid=None,
):
    if kind is None:
        kind = (
            DEVICE_KIND.mouse
            if "Mouse" in name
            else DEVICE_KIND.keyboard
            if "Keyboard" in name
            else DEVICE_KIND.numpad
            if "Number Pad" in name
            else DEVICE_KIND.touchpad
            if "Touchpad" in name
            else DEVICE_KIND.trackball
            if "Trackball" in name
            else None
        )
    assert kind is not None, f"descriptor for {name} does not have kind set"

    if protocol is not None:
        if wpid:
            for w in wpid if isinstance(wpid, tuple) else (wpid,):
                if protocol > 1.0:
                    assert w[0:1] == "4", f"{name} has protocol {protocol:0.1f}, wpid {w}"
                else:
                    if w[0:1] == "1":
                        assert kind == DEVICE_KIND.mouse, f"{name} has protocol {protocol:0.1f}, wpid {w}"
                    elif w[0:1] == "2":
                        assert kind in (
                            DEVICE_KIND.keyboard,
                            DEVICE_KIND.numpad,
                        ), f"{name} has protocol {protocol:0.1f}, wpid {w}"

    device_descriptor = _DeviceDescriptor(
        name=name,
        kind=kind,
        wpid=wpid,
        codename=codename,
        protocol=protocol,
        registers=registers,
        usbid=usbid,
        interface=interface,
        btid=btid,
    )

    if usbid:
        found = get_usbid(usbid)
        assert found is None, f"duplicate usbid in device descriptors: {found}"
    if btid:
        found = get_btid(btid)
        assert found is None, f"duplicate btid in device descriptors: {found}"

    assert codename not in DEVICES, f"duplicate codename in device descriptors: {DEVICES[codename]}"
    if codename:
        DEVICES[codename] = device_descriptor

    if wpid:
        for w in wpid if isinstance(wpid, tuple) else (wpid,):
            assert w not in DEVICES_WPID, f"duplicate wpid in device descriptors: {DEVICES_WPID[w]}"
            DEVICES_WPID[w] = device_descriptor


def get_wpid(wpid):
    return DEVICES_WPID.get(wpid)


def get_codename(codename):
    return DEVICES.get(codename)


def get_usbid(usbid):
    if isinstance(usbid, str):
        usbid = int(usbid, 16)
    found = next((x for x in DEVICES.values() if x.usbid == usbid), None)
    return found


def get_btid(btid):
    if isinstance(btid, str):
        btid = int(btid, 16)
    found = next((x for x in DEVICES.values() if x.btid == btid), None)
    return found


# Some HID++1.0 registers and HID++2.0 features can be discovered at run-time,
# so they are not specified here.
#
# State registers (battery, leds, some features, etc) are only used by
# HID++ 1.0 devices, while HID++ 2.0 devices use features for the same
# functionalities.

# Well-known registers (in hex):
#  * 00 - notification flags (all devices)
#    01 - mice: smooth scrolling
#    07 - battery status
#    09 - keyboards: FN swap (if it has the FN key)
#    0D - battery charge
#       a device may have either the 07 or 0D register available;
#       no known device uses both
#    51 - leds
#    63 - mice: DPI
#  * F1 - firmware info
# Some registers appear to be universally supported, no matter the HID++ version
# (marked with *). The rest may or may not be supported, and their values may or
# may not mean the same thing across different devices.

# The 'codename' and 'kind' fields are usually guessed from the device name,
# but in some cases (like the Logitech Cube) that heuristic fails and they have
# to be specified.
#
# The 'protocol' and 'wpid' fields are optional (they can be discovered at
# runtime), but specifying them here speeds up device discovery and reduces the
# USB traffic Solaar has to do to fully identify peripherals.
#
# The 'registers' field indicates read-only registers, specifying a state. These
# are valid (AFAIK) only to HID++ 1.0 devices.
# The 'settings' field indicates a read/write register; based on them Solaar
# generates, at runtime, the settings controls in the device panel.
# Solaar now sets up this field in settings_templates.py to eliminate a imports loop.
# HID++ 1.0 devices may only have register-based settings; HID++ 2.0 devices may only have
# feature-based settings.

# Devices are organized by kind
# Within kind devices are sorted by wpid, then by usbid, then by btid, with missing values sorted later

# Keyboards

_D("Wireless Keyboard EX110", codename="EX110", protocol=1.0, wpid="0055", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Keyboard S510", codename="S510", protocol=1.0, wpid="0056", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Wave Keyboard K550", codename="K550", protocol=1.0, wpid="0060", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Keyboard EX100", codename="EX100", protocol=1.0, wpid="0065", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Keyboard MK300", codename="MK300", protocol=1.0, wpid="0068", registers=(Reg.BATTERY_STATUS,))
_D("Number Pad N545", codename="N545", protocol=1.0, wpid="2006", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Compact Keyboard K340", codename="K340", protocol=1.0, wpid="2007", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Keyboard MK700", codename="MK700", protocol=1.0, wpid="2008", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Wave Keyboard K350", codename="K350", protocol=1.0, wpid="200A", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Keyboard MK320", codename="MK320", protocol=1.0, wpid="200F", registers=(Reg.BATTERY_STATUS,))
_D(
    "Wireless Illuminated Keyboard K800",
    codename="K800",
    protocol=1.0,
    wpid="2010",
    registers=(Reg.BATTERY_STATUS, Reg.THREE_LEDS),
)
_D("Wireless Keyboard K520", codename="K520", protocol=1.0, wpid="2011", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Solar Keyboard K750", codename="K750", protocol=2.0, wpid="4002")
_D("Wireless Keyboard K270 (unifying)", codename="K270", protocol=2.0, wpid="4003")
_D("Wireless Keyboard K360", codename="K360", protocol=2.0, wpid="4004")
_D("Wireless Keyboard K230", codename="K230", protocol=2.0, wpid="400D")
_D("Wireless Touch Keyboard K400", codename="K400", protocol=2.0, wpid=("400E", "4024"))
_D("Wireless Keyboard MK270", codename="MK270", protocol=2.0, wpid="4023")
_D("Illuminated Living-Room Keyboard K830", codename="K830", protocol=2.0, wpid="4032")
_D("Wireless Touch Keyboard K400 Plus", codename="K400 Plus", protocol=2.0, wpid="404D")
_D("Wireless Multi-Device Keyboard K780", codename="K780", protocol=4.5, wpid="405B")
_D("Wireless Keyboard K375s", codename="K375s", protocol=2.0, wpid="4061")
_D("Craft Advanced Keyboard", codename="Craft", protocol=4.5, wpid="4066", btid=0xB350)
_D("Wireless Illuminated Keyboard K800 new", codename="K800 new", protocol=4.5, wpid="406E")
_D("Wireless Keyboard K470", codename="K470", protocol=4.5, wpid="4075")
_D("MX Keys Keyboard", codename="MX Keys", protocol=4.5, wpid="408A", btid=0xB35B)
_D(
    "G915 TKL LIGHTSPEED Wireless RGB Mechanical Gaming Keyboard",
    codename="G915 TKL",
    protocol=4.2,
    wpid="408E",
    usbid=0xC343,
)
_D("Illuminated Keyboard", codename="Illuminated", protocol=1.0, usbid=0xC318, interface=1)
_D("G213 Prodigy Gaming Keyboard", codename="G213", usbid=0xC336, interface=1)
_D("G512 RGB Mechanical Gaming Keyboard", codename="G512", usbid=0xC33C, interface=1)
_D("G815 Mechanical Keyboard", codename="G815", usbid=0xC33F, interface=1)
_D("diNovo Edge Keyboard", codename="diNovo", protocol=1.0, wpid="C714")
_D("K845 Mechanical Keyboard", codename="K845", usbid=0xC341, interface=3)

# Mice

_D("LX5 Cordless Mouse", codename="LX5", protocol=1.0, wpid="0036", registers=(Reg.BATTERY_STATUS,))
_D("LX7 Cordless Laser Mouse", codename="LX7", protocol=1.0, wpid="0039", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Wave Mouse M550", codename="M550", protocol=1.0, wpid="003C", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Mouse EX100", codename="EX100m", protocol=1.0, wpid="003F", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Mouse M30", codename="M30", protocol=1.0, wpid="0085", registers=(Reg.BATTERY_STATUS,))
_D("MX610 Laser Cordless Mouse", codename="MX610", protocol=1.0, wpid="1001", registers=(Reg.BATTERY_STATUS,))
_D("G7 Cordless Laser Mouse", codename="G7", protocol=1.0, wpid="1002", registers=(Reg.BATTERY_STATUS,))
_D("V400 Laser Cordless Mouse", codename="V400", protocol=1.0, wpid="1003", registers=(Reg.BATTERY_STATUS,))
_D("MX610 Left-Handled Mouse", codename="MX610L", protocol=1.0, wpid="1004", registers=(Reg.BATTERY_STATUS,))
_D("V450 Laser Cordless Mouse", codename="V450", protocol=1.0, wpid="1005", registers=(Reg.BATTERY_STATUS,))
_D(
    "VX Revolution",
    codename="VX Revolution",
    kind=DEVICE_KIND.mouse,
    protocol=1.0,
    wpid=("1006", "100D", "0612"),
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "MX Air",
    codename="MX Air",
    protocol=1.0,
    kind=DEVICE_KIND.mouse,
    wpid=("1007", "100E"),
    registers=Reg.BATTERY_CHARGE,
)
_D(
    "MX Revolution",
    codename="MX Revolution",
    protocol=1.0,
    kind=DEVICE_KIND.mouse,
    wpid=("1008", "100C"),
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "MX620 Laser Cordless Mouse",
    codename="MX620",
    protocol=1.0,
    wpid=("100A", "1016"),
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "VX Nano Cordless Laser Mouse",
    codename="VX Nano",
    protocol=1.0,
    wpid=("100B", "100F"),
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "V450 Nano Cordless Laser Mouse",
    codename="V450 Nano",
    protocol=1.0,
    wpid="1011",
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "V550 Nano Cordless Laser Mouse",
    codename="V550 Nano",
    protocol=1.0,
    wpid="1013",
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "MX 1100 Cordless Laser Mouse",
    codename="MX 1100",
    protocol=1.0,
    kind=DEVICE_KIND.mouse,
    wpid="1014",
    registers=(Reg.BATTERY_CHARGE,),
)
_D("Anywhere Mouse MX", codename="Anywhere MX", protocol=1.0, wpid="1017", registers=(Reg.BATTERY_CHARGE,))
_D(
    "Performance Mouse MX",
    codename="Performance MX",
    protocol=1.0,
    wpid="101A",
    registers=(Reg.BATTERY_STATUS, Reg.THREE_LEDS),
)
_D(
    "Marathon Mouse M705 (M-R0009)",
    codename="M705 (M-R0009)",
    protocol=1.0,
    wpid="101B",
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "Wireless Mouse M350",
    codename="M350",
    protocol=1.0,
    wpid="101C",
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "Wireless Mouse M505",
    codename="M505/B605",
    protocol=1.0,
    wpid="101D",
    registers=(Reg.BATTERY_CHARGE,),
)
_D(
    "Wireless Mouse M305",
    codename="M305",
    protocol=1.0,
    wpid="101F",
    registers=(Reg.BATTERY_STATUS,),
)
_D(
    "Wireless Mouse M215",
    codename="M215",
    protocol=1.0,
    wpid="1020",
)
_D(
    "G700 Gaming Mouse",
    codename="G700",
    protocol=1.0,
    wpid="1023",
    usbid=0xC06B,
    interface=1,
    registers=(
        Reg.BATTERY_STATUS,
        Reg.THREE_LEDS,
    ),
)
_D("Wireless Mouse M310", codename="M310", protocol=1.0, wpid="1024", registers=(Reg.BATTERY_STATUS,))
_D("Wireless Mouse M510", codename="M510", protocol=1.0, wpid="1025", registers=(Reg.BATTERY_STATUS,))
_D("Fujitsu Sonic Mouse", codename="Sonic", protocol=1.0, wpid="1029")
_D(
    "G700s Gaming Mouse",
    codename="G700s",
    protocol=1.0,
    wpid="102A",
    usbid=0xC07C,
    interface=1,
    registers=(
        Reg.BATTERY_STATUS,
        Reg.THREE_LEDS,
    ),
)
_D("Couch Mouse M515", codename="M515", protocol=2.0, wpid="4007")
# _D("Wireless Mouse M175", codename="M175", protocol=2.0, wpid="4008")
_D("Wireless Mouse M325", codename="M325", protocol=2.0, wpid="400A")
_D("Wireless Mouse M525", codename="M525", protocol=2.0, wpid="4013")
_D("Wireless Mouse M345", codename="M345", protocol=2.0, wpid="4017")
_D("Wireless Mouse M187", codename="M187", protocol=2.0, wpid="4019")
_D("Touch Mouse M600", codename="M600", protocol=2.0, wpid="401A")
_D("Wireless Mouse M150", codename="M150", protocol=2.0, wpid="4022")
_D("Wireless Mouse M185", codename="M185", protocol=2.0, wpid="4038")
_D("Wireless Mouse MX Master", codename="MX Master", protocol=4.5, wpid="4041", btid=0xB012)
_D("Anywhere Mouse MX 2", codename="Anywhere MX 2", protocol=4.5, wpid="404A")
_D("Wireless Mouse M510", codename="M510v2", protocol=2.0, wpid="4051")
_D("Wireless Mouse M185 new", codename="M185n", protocol=4.5, wpid="4054")
_D("Wireless Mouse M185/M235/M310", codename="M185/M235/M310", protocol=4.5, wpid="4055")
_D("Wireless Mouse MX Master 2S", codename="MX Master 2S", protocol=4.5, wpid="4069", btid=0xB019)
_D("Multi Device Silent Mouse M585/M590", codename="M585/M590", protocol=4.5, wpid="406B")
_D(
    "Marathon Mouse M705 (M-R0073)",
    codename="M705 (M-R0073)",
    protocol=4.5,
    wpid="406D",
)
_D("MX Vertical Wireless Mouse", codename="MX Vertical", protocol=4.5, wpid="407B", btid=0xB020, usbid=0xC08A)
_D("Wireless Mouse Pebble M350", codename="Pebble", protocol=2.0, wpid="4080")
_D("MX Master 3 Wireless Mouse", codename="MX Master 3", protocol=4.5, wpid="4082", btid=0xB023)
_D("PRO X Wireless", kind="mouse", codename="PRO X", wpid="4093", usbid=0xC094)

_D("G9 Laser Mouse", codename="G9", usbid=0xC048, interface=1, protocol=1.0)
_D("G9x Laser Mouse", codename="G9x", usbid=0xC066, interface=1, protocol=1.0)
_D("G502 Gaming Mouse", codename="G502", usbid=0xC07D, interface=1)
_D("G402 Gaming Mouse", codename="G402", usbid=0xC07E, interface=1)
_D("G900 Chaos Spectrum Gaming Mouse", codename="G900", usbid=0xC081)
_D("G403 Gaming Mouse", codename="G403", usbid=0xC082)
_D("G903 Lightspeed Gaming Mouse", codename="G903", usbid=0xC086)
_D("G703 Lightspeed Gaming Mouse", codename="G703", usbid=0xC087)
_D("GPro Gaming Mouse", codename="GPro", usbid=0xC088)
_D("G502 SE Hero Gaming Mouse", codename="G502 Hero", usbid=0xC08B, interface=1)
_D("G502 Lightspeed Gaming Mouse", codename="G502 Lightspeed", usbid=0xC08D)
_D("MX518 Gaming Mouse", codename="MX518", usbid=0xC08E, interface=1)
_D("G703 Hero Gaming Mouse", codename="G703 Hero", usbid=0xC090)
_D("G903 Hero Gaming Mouse", codename="G903 Hero", usbid=0xC091)
_D(None, kind=DEVICE_KIND.mouse, usbid=0xC092, interface=1)  # two mice share this ID
_D("M500S Mouse", codename="M500S", usbid=0xC093, interface=1)
# _D('G600 Gaming Mouse', codename='G600 Gaming', usbid=0xc24a, interface=1) # not an HID++ device
_D("G500s Gaming Mouse", codename="G500s Gaming", usbid=0xC24E, interface=1, protocol=1.0)
_D("G502 Proteus Spectrum Optical Mouse", codename="G502 Proteus Spectrum", usbid=0xC332, interface=1)
_D("Logitech PRO Gaming Keyboard", codename="PRO Gaming Keyboard", usbid=0xC339, interface=1)

_D("Logitech MX Revolution Mouse M-RCL 124", codename="M-RCL 124", btid=0xB007, interface=1)

# Trackballs

_D("Wireless Trackball M570", codename="M570")

# Touchpads

_D("Wireless Touchpad", codename="Wireless Touch", protocol=2.0, wpid="4011")
_D("Wireless Rechargeable Touchpad T650", codename="T650", protocol=2.0, wpid="4101")
_D(
    "G Powerplay", codename="Powerplay", protocol=2.0, kind=DEVICE_KIND.touchpad, wpid="405F"
)  # To override self-identification

# Headset

_D("G533 Gaming Headset", codename="G533 Headset", protocol=2.0, interface=3, kind=DEVICE_KIND.headset, usbid=0x0A66)
_D("G535 Gaming Headset", codename="G535 Headset", protocol=2.0, interface=3, kind=DEVICE_KIND.headset, usbid=0x0AC4)
_D("G935 Gaming Headset", codename="G935 Headset", protocol=2.0, interface=3, kind=DEVICE_KIND.headset, usbid=0x0A87)
_D("G733 Gaming Headset", codename="G733 Headset", protocol=2.0, interface=3, kind=DEVICE_KIND.headset, usbid=0x0AB5)
_D(
    "G733 Gaming Headset",
    codename="G733 Headset New",
    protocol=2.0,
    interface=3,
    kind=DEVICE_KIND.headset,
    usbid=0x0AFE,
)
_D(
    "PRO X Wireless Gaming Headset",
    codename="PRO Headset",
    protocol=2.0,
    interface=3,
    kind=DEVICE_KIND.headset,
    usbid=0x0ABA,
)



## Copyright (C) Solaar Contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Tuple

from solaar import __version__
from solaar.i18n import _


def _get_current_year() -> int:
    return datetime.now().year


class AboutModel:
    def get_version(self) -> str:
        return __version__

    def get_description(self) -> str:
        return _("Manages Logitech receivers,\nkeyboards, mice, and tablets.")

    def get_copyright(self) -> str:
        return f"© 2012-{_get_current_year()} Daniel Pavel and contributors to the Solaar project"

    def get_authors(self) -> List[str]:
        return [
            "Daniel Pavel http://github.com/pwr",
        ]

    def get_translators(self) -> List[str]:
        return [
            "gogo (croatian)",
            "Papoteur, David Geiger, Damien Lallement (français)",
            "Michele Olivo (italiano)",
            "Adrian Piotrowicz (polski)",
            "Drovetto, JrBenito (Portuguese-BR)",
            "Daniel Pavel (română)",
            "Daniel Zippert, Emelie Snecker (svensk)",
            "Dimitriy Ryazantcev (Russian)",
            "El Jinete Sin Cabeza (Español)",
            "Ferdina Kusumah (Indonesia)",
            "John Erling Blad (Norwegian Bokmål, Norwegian Nynorsk)",
        ]

    def get_credit_sections(self) -> List[Tuple[str, List[str]]]:
        return [
            (_("Additional Programming"), ["Filipe Laíns", "Peter F. Patel-Schneider"]),
            (_("GUI design"), ["Julien Gascard", "Daniel Pavel"]),
            (
                _("Testing"),
                [
                    "Douglas Wagner",
                    "Julien Gascard",
                    "Peter Wu http://www.lekensteyn.nl/logitech-unifying.html",
                ],
            ),
            (
                _("Logitech documentation"),
                [
                    "Julien Danjou http://julien.danjou.info/blog/2012/logitech-unifying-upower",
                    "Nestor Lopez Casado http://drive.google.com/folderview?id=0BxbRzx7vEV7eWmgwazJ3NUFfQ28",
                ],
            ),
        ]

    def get_website(self):
        return "https://pwr-solaar.github.io/Solaar"



## Copyright (C) 2020 Peter Patel-Schneider
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import ctypes
import logging
import math
import numbers
import os
import platform
import socket
import struct
import subprocess
import sys
import time
import typing

from typing import Any
from typing import Dict
from typing import Tuple

import gi
import psutil
import yaml

from keysyms import keysymdef

# There is no evdev on macOS or Windows. Diversion will not work without
# it but other Solaar functionality is available.
if platform.system() in ("Darwin", "Windows"):
    evdev = None
else:
    import evdev

from .common import NamedInt
from .hidpp20 import SupportedFeature
from .special_keys import CONTROL

gi.require_version("Gdk", "3.0")  # isort:skip
from gi.repository import Gdk, GLib  # NOQA: E402 # isort:skip

if typing.TYPE_CHECKING:
    from .base import HIDPPNotification

logger = logging.getLogger(__name__)

#
# See docs/rules.md for documentation
#
# Several capabilities of rules depend on aspects of GDK, X11, or XKB
# As the Solaar GUI uses GTK, Glib and GDK are always available and are obtained from gi.repository
#
# Process condition depends on X11 from python-xlib, and is probably not possible at all in Wayland
# MouseProcess condition depends on X11 from python-xlib, and is probably not possible at all in Wayland
# Modifiers condition depends only on GDK
# KeyPress action determines whether a keysym is a currently-down modifier using get_modifier_mapping from python-xlib;
#   under Wayland no modifier keys are considered down so all modifier keys are pressed, potentially leading to problems
# KeyPress action translates key names to keysysms using the local file described for GUI keyname determination
# KeyPress action gets the current keyboard group using XkbGetState from libX11.so using ctypes definitions
#   under Wayland the keyboard group is None resulting in using the first keyboard group
# KeyPress action translates keysyms to keycodes using the GDK keymap
# KeyPress, MouseScroll, and MouseClick actions use XTest (under X11) or uinput.
# For uinput to work the user must have write access for /dev/uinput.
# To get this access run  sudo setfacl -m u:${user}:rw /dev/uinput
#
# Rule GUI keyname determination uses a local file generated
#   from http://cgit.freedesktop.org/xorg/proto/x11proto/plain/keysymdef.h
#   and http://cgit.freedesktop.org/xorg/proto/x11proto/plain/XF86keysym.h
# because there does not seem to be a non-X11 file for this set of key names

# Setting up is complex because there are several systems that each provide partial facilities:
# GDK - always available (when running with a window system) but only provides access to keymap
# X11 - provides access to active process and process with window under mouse and current modifier keys
# Xtest extension to X11 - provides input simulation, partly works under Wayland
# Wayland - provides input simulation

XK_KEYS: Dict[str, int] = keysymdef.key_symbols

# Event codes - can't use Xlib.X codes because Xlib might not be available
_KEY_RELEASE = 0
_KEY_PRESS = 1
_BUTTON_RELEASE = 2
_BUTTON_PRESS = 3

CLICK, DEPRESS, RELEASE = "click", "depress", "release"

gdisplay = Gdk.Display.get_default()  # can be None if Solaar is run without a full window system
gkeymap = Gdk.Keymap.get_for_display(gdisplay) if gdisplay else None
if logger.isEnabledFor(logging.INFO):
    logger.info("GDK Keymap %sset up", "" if gkeymap else "not ")

wayland = os.getenv("WAYLAND_DISPLAY")  # is this Wayland?
if wayland:
    logger.warning(
        "rules cannot access modifier keys in Wayland, "
        "accessing process only works on GNOME with Solaar Gnome extension installed"
    )

try:
    import Xlib

    _x11 = None  # X11 might be available
except Exception:
    _x11 = False  # X11 is not available

# Globals
xtest_available = True  # Xtest might be available
xdisplay = None


Xkbdisplay = None  # xkb might be available
X11Lib = None

modifier_keycodes = []
XkbUseCoreKbd = 0x100
NET_ACTIVE_WINDOW = None
NET_WM_PID = None
WM_CLASS = None


udevice = None

key_down = None
key_up = None

keys_down = []
g_keys_down = 0
m_keys_down = 0
mr_key_down = False
thumb_wheel_displacement = 0

_dbus_interface = None


class XkbDisplay(ctypes.Structure):
    """opaque struct"""


class XkbStateRec(ctypes.Structure):
    _fields_ = [
        ("group", ctypes.c_ubyte),
        ("locked_group", ctypes.c_ubyte),
        ("base_group", ctypes.c_ushort),
        ("latched_group", ctypes.c_ushort),
        ("mods", ctypes.c_ubyte),
        ("base_mods", ctypes.c_ubyte),
        ("latched_mods", ctypes.c_ubyte),
        ("locked_mods", ctypes.c_ubyte),
        ("compat_state", ctypes.c_ubyte),
        ("grab_mods", ctypes.c_ubyte),
        ("compat_grab_mods", ctypes.c_ubyte),
        ("lookup_mods", ctypes.c_ubyte),
        ("compat_lookup_mods", ctypes.c_ubyte),
        ("ptr_buttons", ctypes.c_ushort),
    ]  # something strange is happening here but it is not being used


def x11_setup():
    global _x11, xdisplay, modifier_keycodes, NET_ACTIVE_WINDOW, NET_WM_PID, WM_CLASS, xtest_available
    if _x11 is not None:
        return _x11
    try:
        from Xlib.display import Display

        xdisplay = Display()
        modifier_keycodes = xdisplay.get_modifier_mapping()  # there should be a way to do this in Gdk
        NET_ACTIVE_WINDOW = xdisplay.intern_atom("_NET_ACTIVE_WINDOW")
        NET_WM_PID = xdisplay.intern_atom("_NET_WM_PID")
        WM_CLASS = xdisplay.intern_atom("WM_CLASS")
        _x11 = True  # X11 available
        if logger.isEnabledFor(logging.INFO):
            logger.info("X11 library loaded and display set up")
    except Exception:
        logger.warning("X11 not available - some rule capabilities inoperable", exc_info=sys.exc_info())
        _x11 = False
        xtest_available = False
    return _x11


def gnome_dbus_interface_setup():
    global _dbus_interface
    if _dbus_interface is not None:
        return _dbus_interface
    try:
        import dbus

        bus = dbus.SessionBus()
        remote_object = bus.get_object("org.gnome.Shell", "/io/github/pwr_solaar/solaar")
        _dbus_interface = dbus.Interface(remote_object, "io.github.pwr_solaar.solaar")
    except dbus.exceptions.DBusException:
        logger.warning(
            "Solaar Gnome extension not installed - some rule capabilities inoperable",
            exc_info=sys.exc_info(),
        )
        _dbus_interface = False
    return _dbus_interface


def xkb_setup():
    global X11Lib, Xkbdisplay
    if Xkbdisplay is not None:
        return Xkbdisplay
    try:  # set up to get keyboard state using ctypes interface to libx11
        X11Lib = ctypes.cdll.LoadLibrary("libX11.so")
        X11Lib.XOpenDisplay.restype = ctypes.POINTER(XkbDisplay)
        X11Lib.XkbGetState.argtypes = [ctypes.POINTER(XkbDisplay), ctypes.c_uint, ctypes.POINTER(XkbStateRec)]
        Xkbdisplay = X11Lib.XOpenDisplay(None)
        if logger.isEnabledFor(logging.INFO):
            logger.info("XKB display set up")
    except Exception:
        logger.warning("XKB display not available - rules cannot access keyboard group", exc_info=sys.exc_info())
        Xkbdisplay = False
    return Xkbdisplay


if evdev:
    buttons = {
        "unknown": (None, None),
        "left": (1, evdev.ecodes.ecodes["BTN_LEFT"]),
        "middle": (2, evdev.ecodes.ecodes["BTN_MIDDLE"]),
        "right": (3, evdev.ecodes.ecodes["BTN_RIGHT"]),
        "scroll_up": (4, evdev.ecodes.ecodes["BTN_4"]),
        "scroll_down": (5, evdev.ecodes.ecodes["BTN_5"]),
        "scroll_left": (6, evdev.ecodes.ecodes["BTN_6"]),
        "scroll_right": (7, evdev.ecodes.ecodes["BTN_7"]),
        "button8": (8, evdev.ecodes.ecodes["BTN_8"]),
        "button9": (9, evdev.ecodes.ecodes["BTN_9"]),
        "back": (10, evdev.ecodes.ecodes["BTN_SIDE"]),
        "forward": (11, evdev.ecodes.ecodes["BTN_EXTRA"]),
    }

    # uinput capability for keyboard keys, mouse buttons, and scrolling
    key_events = [c for n, c in evdev.ecodes.ecodes.items() if n.startswith("KEY") and n != "KEY_CNT"]
    for _, evcode in buttons.values():
        if evcode:
            key_events.append(evcode)
    devicecap = {
        evdev.ecodes.EV_KEY: key_events,
        evdev.ecodes.EV_REL: [evdev.ecodes.REL_WHEEL, evdev.ecodes.REL_HWHEEL],
    }
else:
    # Just mock these since they won't be useful without evdev anyway
    buttons = {}
    key_events = []
    devicecap = {}


def setup_uinput():
    global udevice
    if udevice is not None:
        return udevice
    try:
        udevice = evdev.uinput.UInput(events=devicecap, name="solaar-keyboard")
        if logger.isEnabledFor(logging.INFO):
            logger.info("uinput device set up")
        return True
    except Exception as e:
        logger.warning("cannot create uinput device: %s", e)


if wayland:  # Wayland can't use xtest so may as well set up uinput now
    setup_uinput()


def kbdgroup():
    if xkb_setup():
        state = XkbStateRec()
        X11Lib.XkbGetState(Xkbdisplay, XkbUseCoreKbd, ctypes.pointer(state))
        return state.group
    else:
        return None


def modifier_code(keycode):
    if wayland or not x11_setup() or keycode == 0:
        return None
    for m in range(0, len(modifier_keycodes)):
        if keycode in modifier_keycodes[m]:
            return m


def signed(bytes_: bytes) -> int:
    return int.from_bytes(bytes_, "big", signed=True)


def xy_direction(_x, _y):
    # normalize x and y
    m = math.sqrt((_x * _x) + (_y * _y))
    if m == 0:
        return "noop"
    x = round(_x / m)
    y = round(_y / m)
    if x < 0 and y < 0:
        return "Mouse Up-left"
    elif x > 0 > y:
        return "Mouse Up-right"
    elif x < 0 < y:
        return "Mouse Down-left"
    elif x > 0 and y > 0:
        return "Mouse Down-right"
    elif x > 0:
        return "Mouse Right"
    elif x < 0:
        return "Mouse Left"
    elif y > 0:
        return "Mouse Down"
    elif y < 0:
        return "Mouse Up"
    else:
        return "noop"


def simulate_xtest(code, event):
    global xtest_available
    if x11_setup() and xtest_available:
        try:
            event = (
                Xlib.X.KeyPress
                if event == _KEY_PRESS
                else Xlib.X.KeyRelease
                if event == _KEY_RELEASE
                else Xlib.X.ButtonPress
                if event == _BUTTON_PRESS
                else Xlib.X.ButtonRelease
                if event == _BUTTON_RELEASE
                else None
            )
            Xlib.ext.xtest.fake_input(xdisplay, event, code)
            xdisplay.sync()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("xtest simulated input %s %s %s", xdisplay, event, code)
            return True
        except Exception as e:
            xtest_available = False
            logger.warning("xtest fake input failed: %s", e)


def simulate_uinput(what, code, arg):
    global udevice
    if setup_uinput():
        try:
            udevice.write(what, code, arg)
            udevice.syn()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("uinput simulated input %s %s %s", what, code, arg)
            return True
        except Exception as e:
            udevice = None
            logger.warning("uinput write failed: %s", e)


def simulate_key(code, event):  # X11 keycode but Solaar event code
    if not wayland and simulate_xtest(code, event):
        return True
    if evdev and simulate_uinput(evdev.ecodes.EV_KEY, code - 8, event):
        return True
    logger.warning("no way to simulate key input")


def click_xtest(button, count):
    if isinstance(count, int):
        for _ in range(count):
            if not simulate_xtest(button[0], _BUTTON_PRESS):
                return False
            if not simulate_xtest(button[0], _BUTTON_RELEASE):
                return False
    else:
        if count != RELEASE:
            if not simulate_xtest(button[0], _BUTTON_PRESS):
                return False
        if count != DEPRESS:
            if not simulate_xtest(button[0], _BUTTON_RELEASE):
                return False
    return True


def click_uinput(button, count):
    if isinstance(count, int):
        for _ in range(count):
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 1):
                return False
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 0):
                return False
    else:
        if count != RELEASE:
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 1):
                return False
        if count != DEPRESS:
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 0):
                return False
    return True


def click(button, count):
    if not wayland and click_xtest(button, count):
        return True
    if click_uinput(button, count):
        return True
    logger.warning("no way to simulate mouse click")
    return False


def simulate_scroll(dx, dy):
    if not wayland and xtest_available:
        success = True
        if dx:
            success = click_xtest(buttons["scroll_right" if dx > 0 else "scroll_left"], count=abs(dx))
        if dy and success:
            success = click_xtest(buttons["scroll_up" if dy > 0 else "scroll_down"], count=abs(dy))
        if success:
            return True
    if setup_uinput():
        success = True
        if dx:
            success = simulate_uinput(evdev.ecodes.EV_REL, evdev.ecodes.REL_HWHEEL, dx)
        if dy and success:
            success = simulate_uinput(evdev.ecodes.EV_REL, evdev.ecodes.REL_WHEEL, dy)
        if success:
            return True
    logger.warning("no way to simulate scrolling")


def thumb_wheel_up(f, r, d, a):
    global thumb_wheel_displacement
    if f != SupportedFeature.THUMB_WHEEL or r != 0:
        return False
    if a is None:
        return signed(d[0:2]) < 0 and signed(d[0:2])
    elif thumb_wheel_displacement <= -a:
        thumb_wheel_displacement += a
        return 1
    else:
        return False


def thumb_wheel_down(f, r, d, a):
    global thumb_wheel_displacement
    if f != SupportedFeature.THUMB_WHEEL or r != 0:
        return False
    if a is None:
        return signed(d[0:2]) > 0 and signed(d[0:2])
    elif thumb_wheel_displacement >= a:
        thumb_wheel_displacement -= a
        return 1
    else:
        return False


def charging(f, r, d, _a):
    if (
        (f == SupportedFeature.BATTERY_STATUS and r == 0 and 1 <= d[2] <= 4)
        or (f == SupportedFeature.BATTERY_VOLTAGE and r == 0 and d[2] & (1 << 7))
        or (f == SupportedFeature.UNIFIED_BATTERY and r == 0 and 1 <= d[2] <= 3)
    ):
        return 1
    else:
        return False


TESTS = {
    "crown_right": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[1] < 128 and d[1], False],
    "crown_left": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[1] >= 128 and 256 - d[1], False],
    "crown_right_ratchet": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[2] < 128 and d[2], False],
    "crown_left_ratchet": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[2] >= 128 and 256 - d[2], False],
    "crown_tap": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[5] == 0x01 and d[5], False],
    "crown_start_press": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[6] == 0x01 and d[6], False],
    "crown_end_press": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and d[6] == 0x05 and d[6], False],
    "crown_pressed": [lambda f, r, d, a: f == SupportedFeature.CROWN and r == 0 and 0x01 <= d[6] <= 0x04 and d[6], False],
    "thumb_wheel_up": [thumb_wheel_up, True],
    "thumb_wheel_down": [thumb_wheel_down, True],
    "lowres_wheel_up": [
        lambda f, r, d, a: f == SupportedFeature.LOWRES_WHEEL and r == 0 and signed(d[0:1]) > 0 and signed(d[0:1]),
        False,
    ],
    "lowres_wheel_down": [
        lambda f, r, d, a: f == SupportedFeature.LOWRES_WHEEL and r == 0 and signed(d[0:1]) < 0 and signed(d[0:1]),
        False,
    ],
    "hires_wheel_up": [
        lambda f, r, d, a: f == SupportedFeature.HIRES_WHEEL and r == 0 and signed(d[1:3]) > 0 and signed(d[1:3]),
        False,
    ],
    "hires_wheel_down": [
        lambda f, r, d, a: f == SupportedFeature.HIRES_WHEEL and r == 0 and signed(d[1:3]) < 0 and signed(d[1:3]),
        False,
    ],
    "charging": [charging, False],
    "False": [lambda f, r, d, a: False, False],
    "True": [lambda f, r, d, a: True, False],
}

MOUSE_GESTURE_TESTS = {
    "mouse-down": ["Mouse Down"],
    "mouse-up": ["Mouse Up"],
    "mouse-left": ["Mouse Left"],
    "mouse-right": ["Mouse Right"],
    "mouse-noop": [],
}

# COMPONENTS = {}


class RuleComponent:
    def compile(self, c):
        if isinstance(c, RuleComponent):
            return c
        elif isinstance(c, dict) and len(c) == 1:
            k, v = next(iter(c.items()))
            if k in COMPONENTS:
                return COMPONENTS[k](v)
        logger.warning("illegal component in rule: %s", c)
        return Condition()


def _evaluate(components, feature, notification: HIDPPNotification, device, result) -> Any:
    res = True
    for component in components:
        res = component.evaluate(feature, notification, device, result)
        if not isinstance(component, Action) and res is None:
            return None
        if isinstance(component, Condition) and not res:
            return res
    return res


class Rule(RuleComponent):
    def __init__(self, args, source=None, warn=True):
        self.components = [self.compile(a) for a in args]
        self.source = source

    def __str__(self):
        source = f"({self.source})" if self.source else ""
        return f"Rule{source}[{', '.join([c.__str__() for c in self.components])}]"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate rule: %s", self)
        return _evaluate(self.components, feature, notification, device, True)

    def once(self, feature, notification: HIDPPNotification, device, last_result):
        self.evaluate(feature, notification, device, last_result)
        return False

    def data(self):
        return {"Rule": [c.data() for c in self.components]}


class Condition(RuleComponent):
    def __init__(self, *args):
        pass

    def __str__(self):
        return "CONDITION"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return False


class Not(Condition):
    def __init__(self, op, warn=True):
        if isinstance(op, list) and len(op) == 1:
            op = op[0]
        self.op = op
        self.component = self.compile(op)

    def __str__(self):
        return f"Not: {str(self.component)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        result = self.component.evaluate(feature, notification, device, last_result)
        return None if result is None else not result

    def data(self):
        return {"Not": self.component.data()}


class Or(Condition):
    def __init__(self, args, warn=True):
        self.components = [self.compile(a) for a in args]

    def __str__(self):
        return "Or: [" + ", ".join(str(c) for c in self.components) + "]"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        result = False
        for component in self.components:
            result = component.evaluate(feature, notification, device, last_result)
            if not isinstance(component, Action) and result is None:
                return None
            if isinstance(component, Condition) and result:
                return result
        return result

    def data(self):
        return {"Or": [c.data() for c in self.components]}


class And(Condition):
    def __init__(self, args, warn=True):
        self.components = [self.compile(a) for a in args]

    def __str__(self):
        return "And: [" + ", ".join(str(c) for c in self.components) + "]"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return _evaluate(self.components, feature, notification, device, last_result)

    def data(self):
        return {"And": [c.data() for c in self.components]}


def x11_focus_prog():
    if not x11_setup():
        return None
    pid = wm_class = None
    window = xdisplay.get_input_focus().focus
    while window:
        pid = window.get_full_property(NET_WM_PID, 0)
        wm_class = window.get_wm_class()
        if wm_class and pid:
            break
        window = window.query_tree().parent
    try:
        name = psutil.Process(pid.value[0]).name() if pid else ""
    except Exception:
        name = ""
    return (wm_class[0], wm_class[1], name) if wm_class else (name,)


def x11_pointer_prog():
    if not x11_setup():
        return None
    pid = wm_class = None
    window = xdisplay.screen().root.query_pointer().child
    for child in reversed(window.query_tree().children):
        pid = child.get_full_property(NET_WM_PID, 0)
        wm_class = child.get_wm_class()
        if wm_class:
            break
    name = psutil.Process(pid.value[0]).name() if pid else ""
    return (wm_class[0], wm_class[1], name) if wm_class else (name,)


def gnome_dbus_focus_prog():
    if not gnome_dbus_interface_setup():
        return None
    wm_class = _dbus_interface.ActiveWindow()
    return (wm_class,) if wm_class else None


def gnome_dbus_pointer_prog():
    if not gnome_dbus_interface_setup():
        return None
    wm_class = _dbus_interface.PointerOverWindow()
    return (wm_class,) if wm_class else None


class Process(Condition):
    def __init__(self, process, warn=True):
        self.process = process
        if (not wayland and not x11_setup()) or (wayland and not gnome_dbus_interface_setup()):
            if warn:
                logger.warning(
                    "rules can only access active process in X11 or in Wayland under GNOME with Solaar Gnome "
                    "extension - %s",
                    self,
                )
        if not isinstance(process, str):
            if warn:
                logger.warning("rule Process argument not a string: %s", process)
            self.process = str(process)

    def __str__(self):
        return f"Process: {str(self.process)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        if not isinstance(self.process, str):
            return False
        focus = x11_focus_prog() if not wayland else gnome_dbus_focus_prog()
        result = any(bool(s and s.startswith(self.process)) for s in focus) if focus else None
        return result

    def data(self):
        return {"Process": str(self.process)}


class MouseProcess(Condition):
    def __init__(self, process, warn=True):
        self.process = process
        if (not wayland and not x11_setup()) or (wayland and not gnome_dbus_interface_setup()):
            if warn:
                logger.warning(
                    "rules cannot access active mouse process "
                    "in X11 or in Wayland under GNOME with Solaar Extension for GNOME - %s",
                    self,
                )
        if not isinstance(process, str):
            if warn:
                logger.warning("rule MouseProcess argument not a string: %s", process)
            self.process = str(process)

    def __str__(self):
        return f"MouseProcess: {str(self.process)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        if not isinstance(self.process, str):
            return False
        pointer_focus = x11_pointer_prog() if not wayland else gnome_dbus_pointer_prog()
        result = any(bool(s and s.startswith(self.process)) for s in pointer_focus) if pointer_focus else None
        return result

    def data(self):
        return {"MouseProcess": str(self.process)}


class Feature(Condition):
    def __init__(self, feature: str, warn: bool = True):
        try:
            self.feature = SupportedFeature[feature.replace(" ", "_")]
        except KeyError:
            self.feature = None
            if warn:
                logger.warning("rule Feature argument not name of a feature: %s", feature)

    def __str__(self):
        return f"Feature: {str(self.feature)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return feature == self.feature

    def data(self):
        return {"Feature": str(self.feature)}


class Report(Condition):
    def __init__(self, report, warn=True):
        if not (isinstance(report, int)):
            if warn:
                logger.warning("rule Report argument not an integer: %s", report)
            self.report = -1
        else:
            self.report = report

    def __str__(self):
        return f"Report: {str(self.report)}"

    def evaluate(self, report, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return (notification.address >> 4) == self.report

    def data(self):
        return {"Report": self.report}


# Setting(device, setting, [key], value...)
class Setting(Condition):
    def __init__(self, args, warn=True):
        if not (isinstance(args, list) and len(args) > 2):
            if warn:
                logger.warning("rule Setting argument not list with minimum length 3: %s", args)
            self.args = []
        else:
            self.args = args

    def __str__(self):
        return "Setting: " + " ".join([str(a) for a in self.args])

    def evaluate(self, report, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        if len(self.args) < 3:
            return None
        dev = device.find(self.args[0]) if self.args[0] is not None else device
        if dev is None:
            logger.warning("Setting condition: device %s is not known", self.args[0])
            return False
        setting = next((s for s in dev.settings if s.name == self.args[1]), None)
        if setting is None:
            logger.warning("Setting condition: setting %s is not the name of a setting for %s", self.args[1], dev.name)
            return None
        # should the value argument be checked to be sure it is acceptable?? needs to be careful about boolean toggle
        # TODO add compare  methods for more validators
        try:
            result = setting.compare(self.args[2:], setting.read())
        except Exception as e:
            logger.warning("Setting condition: error when checking setting %s: %s", self.args, e)
            result = False
        return result

    def data(self):
        return {"Setting": self.args[:]}


MODIFIERS = {
    "Shift": int(Gdk.ModifierType.SHIFT_MASK),
    "Control": int(Gdk.ModifierType.CONTROL_MASK),
    "Alt": int(Gdk.ModifierType.MOD1_MASK),
    "Super": int(Gdk.ModifierType.MOD4_MASK),
}
MODIFIER_MASK = MODIFIERS["Shift"] + MODIFIERS["Control"] + MODIFIERS["Alt"] + MODIFIERS["Super"]


class Modifiers(Condition):
    def __init__(self, modifiers, warn=True):
        modifiers = [modifiers] if isinstance(modifiers, str) else modifiers
        self.desired = 0
        self.modifiers = []
        for k in modifiers:
            if k in MODIFIERS:
                self.desired += MODIFIERS.get(k, 0)
                self.modifiers.append(k)
            else:
                if warn:
                    logger.warning("unknown rule Modifier value: %s", k)

    def __str__(self):
        return f"Modifiers: {str(self.desired)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        if gkeymap:
            current = gkeymap.get_modifier_state()  # get the current keyboard modifier
            return self.desired == (current & MODIFIER_MASK)
        else:
            logger.warning("no keymap so cannot determine modifier keys")
            return False

    def data(self):
        return {"Modifiers": [str(m) for m in self.modifiers]}


class Key(Condition):
    DOWN = "pressed"
    UP = "released"

    def __init__(self, args, warn=True):
        default_key = 0
        default_action = self.DOWN

        key, action = None, None

        if not args or not isinstance(args, (list, str)):
            if warn:
                logger.warning(f"rule Key arguments unknown: {args}")
            key = default_key
            action = default_action
        elif isinstance(args, str):
            logger.debug(f'rule Key assuming action "{default_action}" for "{args}"')
            key = args
            action = default_action
        elif isinstance(args, list):
            if len(args) == 1:
                logger.debug(f'rule Key assuming action "{default_action}" for "{args}"')
                key, action = args[0], default_action
            elif len(args) >= 2:
                key, action = args[:2]

        if isinstance(key, str) and key in CONTROL:
            self.key = CONTROL[key]
        elif isinstance(key, str) and key.startswith("unknown:"):
            logger.info(f"rule Key key name currently unknown: {key}")
            self.key = CONTROL[int(key[-4:], 16)]
        else:
            if warn:
                logger.warning(f"rule Key key name not name of a Logitech key: {key}")
            self.key = default_key

        if isinstance(action, str) and action in (self.DOWN, self.UP):
            self.action = action
        else:
            if warn:
                logger.warning(f"rule Key action unknown: {action}, assuming {default_action}")
            self.action = default_action

    def __str__(self):
        return f"Key: {str(self.key) if self.key else 'None'} ({self.action})"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return bool(self.key and self.key == (key_down if self.action == self.DOWN else key_up))

    def data(self):
        return {"Key": [str(self.key), self.action]}


class KeyIsDown(Condition):
    def __init__(self, args, warn=True):
        default_key = 0

        key = None

        if not args or not isinstance(args, str):
            if warn:
                logger.warning(f"rule KeyDown arguments unknown: {args}")
            key = default_key
        elif isinstance(args, str):
            key = args

        if isinstance(key, str) and key in CONTROL:
            self.key = CONTROL[key]
        else:
            if warn:
                logger.warning(f"rule Key key name not name of a Logitech key: {key}")
            self.key = default_key

    def __str__(self):
        return f"KeyIsDown: {str(self.key) if self.key else 'None'}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return key_is_down(self.key)

    def data(self):
        return {"KeyIsDown": str(self.key)}


def bit_test(start, end, bits):
    return lambda f, r, d: int.from_bytes(d[start:end], byteorder="big", signed=True) & bits


def range_test(start, end, min, max):
    def range_test_helper(_f, _r, d):
        value = int.from_bytes(d[start:end], byteorder="big", signed=True)
        return min <= value <= max and (value if value else True)

    return range_test_helper


class Test(Condition):
    def __init__(self, test, warn=True):
        self.test = ""
        self.parameter = None
        if isinstance(test, str):
            test = [test]
        if isinstance(test, list) and all(isinstance(t, int) for t in test):
            if warn:
                logger.warning("Test rules consisting of numbers are deprecated, converting to a TestBytes condition")
            self.__class__ = TestBytes
            self.__init__(test, warn=warn)
        elif isinstance(test, list):
            if test[0] in MOUSE_GESTURE_TESTS:
                if warn:
                    logger.warning("mouse movement test %s deprecated, converting to a MouseGesture", test)
                self.__class__ = MouseGesture
                self.__init__(MOUSE_GESTURE_TESTS[0][test], warn=warn)
            elif test[0] in TESTS:
                self.test = test[0]
                self.function = TESTS[test[0]][0]
                self.parameter = test[1] if len(test) > 1 else None
            else:
                if warn:
                    logger.warning("rule Test name not name of a test: %s", test)
                self.test = "False"
                self.function = TESTS["False"][0]
        else:
            if warn:
                logger.warning("rule Test argument not valid %s", test)

    def __str__(self):
        return f"Test: {str(self.test)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return self.function(feature, notification.address, notification.data, self.parameter)

    def data(self):
        return {"Test": ([self.test, self.parameter] if self.parameter is not None else [self.test])}


class TestBytes(Condition):
    def __init__(self, test, warn=True):
        self.test = test
        if (
            isinstance(test, list)
            and 2 < len(test) <= 4
            and all(isinstance(t, int) for t in test)
            and 0 <= test[0] <= 16
            and 0 <= test[1] <= 16
            and test[0] < test[1]
        ):
            self.function = bit_test(*test) if len(test) == 3 else range_test(*test)
        else:
            if warn:
                logger.warning("rule TestBytes argument not valid %s", test)

    def __str__(self):
        return f"TestBytes: {str(self.test)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return self.function(feature, notification.address, notification.data)

    def data(self):
        return {"TestBytes": self.test[:]}


class MouseGesture(Condition):
    MOVEMENTS = [
        "Mouse Up",
        "Mouse Down",
        "Mouse Left",
        "Mouse Right",
        "Mouse Up-left",
        "Mouse Up-right",
        "Mouse Down-left",
        "Mouse Down-right",
    ]

    def __init__(self, movements, warn=True):
        if isinstance(movements, str):
            movements = [movements]
        for x in movements:
            if x not in self.MOVEMENTS and x not in CONTROL:
                if warn:
                    logger.warning("rule Mouse Gesture argument not direction or name of a Logitech key: %s", x)
        self.movements = movements

    def __str__(self):
        return "MouseGesture: " + " ".join(self.movements)

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        if feature == SupportedFeature.MOUSE_GESTURE:
            d = notification.data
            data = struct.unpack("!" + (int(len(d) / 2) * "h"), d)
            data_offset = 1
            movement_offset = 0
            if self.movements and self.movements[0] not in self.MOVEMENTS:  # matching against initiating key
                movement_offset = 1
                if self.movements[0] != str(CONTROL[data[0]]):
                    return False
            for m in self.movements[movement_offset:]:
                if data_offset >= len(data):
                    return False
                if data[data_offset] == 0:
                    direction = xy_direction(data[data_offset + 1], data[data_offset + 2])
                    if m != direction:
                        return False
                    data_offset += 3
                elif data[data_offset] == 1:
                    if m != str(CONTROL[data[data_offset + 1]]):
                        return False
                    data_offset += 2
            return data_offset == len(data)
        return False

    def data(self):
        return {"MouseGesture": [str(m) for m in self.movements]}


class Active(Condition):
    def __init__(self, devID, warn=True):
        if not (isinstance(devID, str)):
            if warn:
                logger.warning("rule Active argument not a string: %s", devID)
            self.devID = ""
        self.devID = devID

    def __str__(self):
        return f"Active: {str(self.devID)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        dev = device.find(self.devID)
        return bool(dev and dev.ping())

    def data(self):
        return {"Active": self.devID}


class Device(Condition):
    def __init__(self, devID, warn=True):
        if not (isinstance(devID, str)):
            if warn:
                logger.warning("rule Device argument not a string: %s", devID)
            self.devID = ""
        self.devID = devID

    def __str__(self):
        return f"Device: {str(self.devID)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        return (
            device.unitId == self.devID
            or device.serial == self.devID
            or device.codename == self.devID
            or device.name == self.devID
        )

    def data(self):
        return {"Device": self.devID}


class Host(Condition):
    def __init__(self, host, warn=True):
        if not (isinstance(host, str)):
            if warn:
                logger.warning("rule Host Name argument not a string: %s", host)
            self.host = ""
        self.host = host

    def __str__(self):
        return f"Host: {str(self.host)}"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("evaluate condition: %s", self)
        hostname = socket.getfqdn()
        return hostname.startswith(self.host)

    def data(self):
        return {"Host": self.host}


class Action(RuleComponent):
    def __init__(self, *args):
        pass

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        return None


def keysym_to_keycode(keysym, _modifiers) -> Tuple[int, int]:  # maybe should take shift into account
    """Reverse the keycode to keysym mapping.

    Warning:
    This is an attempt to reverse the keycode to keysym mappping in XKB.
    It may not be completely general.
    """
    group = kbdgroup() or 0
    keycodes = gkeymap.get_entries_for_keyval(keysym)
    (keycode, level) = (None, None)
    for k in keycodes.keys:  # mappings that have the correct group
        if group == k.group and k.keycode < 256 and (level is None or k.level < level):
            keycode = k.keycode
            level = k.level
    if keycode or group == 0:
        return keycode, level

    for k in keycodes.keys:  # mappings for group 0 where keycode only has group 0 mappings
        if 0 == k.group and k.keycode < 256 and (level is None or k.level < level):
            (a, m, vs) = gkeymap.get_entries_for_keycode(k.keycode)
            if a and all(mk.group == 0 for mk in m):
                keycode = k.keycode
                level = k.level
    return keycode, level


class KeyPress(Action):
    def __init__(self, args, warn=True):
        self.key_names, self.action = self.regularize_args(args)
        if not isinstance(self.key_names, list):
            if warn:
                logger.warning("rule KeyPress keys not key names %s", self.keys_names)
            self.key_symbols = []
        else:
            self.key_symbols = [XK_KEYS.get(k, None) for k in self.key_names]
        if not all(self.key_symbols):
            if warn:
                logger.warning("rule KeyPress keys not key names %s", self.key_names)
            self.key_symbols = []

    def regularize_args(self, args):
        action = CLICK
        if not isinstance(args, list):
            args = [args]
        keys = args
        if len(args) == 2 and args[1] in [CLICK, DEPRESS, RELEASE]:
            keys = [args[0]] if isinstance(args[0], str) else args[0]
            action = args[1]
        return keys, action

    def __str__(self):
        return "KeyPress: " + " ".join(self.key_names) + " " + self.action

    def needed(self, k, modifiers):
        code = modifier_code(k)
        return not (code is not None and modifiers & (1 << code))

    def mods(self, level, modifiers, direction):
        if level == 2 or level == 3:
            (sk, _) = keysym_to_keycode(XK_KEYS.get("ISO_Level3_Shift", None), modifiers)
            if sk and self.needed(sk, modifiers):
                simulate_key(sk, direction)
        if level == 1 or level == 3:
            (sk, _) = keysym_to_keycode(XK_KEYS.get("Shift_L", None), modifiers)
            if sk and self.needed(sk, modifiers):
                simulate_key(sk, direction)

    def keyDown(self, keysyms_, modifiers):
        for k in keysyms_:
            (keycode, level) = keysym_to_keycode(k, modifiers)
            if keycode is None:
                logger.warning("rule KeyPress key symbol not currently available %s", self)
            elif self.action != CLICK or self.needed(keycode, modifiers):  # only check needed when clicking
                self.mods(level, modifiers, _KEY_PRESS)
                simulate_key(keycode, _KEY_PRESS)

    def keyUp(self, keysyms_, modifiers):
        for k in keysyms_:
            (keycode, level) = keysym_to_keycode(k, modifiers)
            if keycode and (self.action != CLICK or self.needed(keycode, modifiers)):  # only check needed when clicking
                simulate_key(keycode, _KEY_RELEASE)
                self.mods(level, modifiers, _KEY_RELEASE)

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if gkeymap:
            current = gkeymap.get_modifier_state()
            if logger.isEnabledFor(logging.INFO):
                logger.info(
                    "KeyPress action: %s %s, group %s, modifiers %s",
                    self.key_names,
                    self.action,
                    kbdgroup(),
                    current,
                )
            if self.action != RELEASE:
                self.keyDown(self.key_symbols, current)
            if self.action != DEPRESS:
                self.keyUp(reversed(self.key_symbols), current)
            time.sleep(0.01)
        else:
            logger.warning("no keymap so cannot determine which keycode to send")
        return None

    def data(self):
        return {"KeyPress": [[str(k) for k in self.key_names], self.action]}


# KeyDown is dangerous as the key can auto-repeat and make your system unusable
# class KeyDown(KeyPress):
#    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
#        super().keyDown(self.keys, current_key_modifiers)
# class KeyUp(KeyPress):
#    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
#        super().keyUp(self.keys, current_key_modifiers)


class MouseScroll(Action):
    def __init__(self, amounts, warn=True):
        if len(amounts) == 1 and isinstance(amounts[0], list):
            amounts = amounts[0]
        if not (len(amounts) == 2 and all([isinstance(a, numbers.Number) for a in amounts])):
            if warn:
                logger.warning("rule MouseScroll argument not two numbers %s", amounts)
            amounts = [0, 0]
        self.amounts = amounts

    def __str__(self):
        return "MouseScroll: " + " ".join([str(a) for a in self.amounts])

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        amounts = self.amounts
        if isinstance(last_result, numbers.Number):
            amounts = [math.floor(last_result * a) for a in self.amounts]
        if logger.isEnabledFor(logging.INFO):
            logger.info("MouseScroll action: %s %s %s", self.amounts, last_result, amounts)
        dx, dy = amounts
        simulate_scroll(dx, dy)
        time.sleep(0.01)
        return None

    def data(self):
        return {"MouseScroll": self.amounts[:]}


class MouseClick(Action):
    def __init__(self, args, warn=True):
        if len(args) == 1 and isinstance(args[0], list):
            args = args[0]
        if not isinstance(args, list):
            args = [args]
        self.button = str(args[0]) if len(args) >= 0 else None
        if self.button not in buttons:
            if warn:
                logger.warning("rule MouseClick action: button %s not known", self.button)
            self.button = None
        count = args[1] if len(args) >= 2 else 1
        try:
            self.count = int(count)
        except (ValueError, TypeError):
            if count in [CLICK, DEPRESS, RELEASE]:
                self.count = count
            elif warn:
                logger.warning(
                    "rule MouseClick action: argument %s should be an integer or click, depress, or release",
                    count,
                )
                self.count = 1

    def __str__(self):
        return f"MouseClick: {self.button} ({str(self.count)})"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.INFO):
            logger.info(f"MouseClick action: {str(self.count)} {self.button}")
        if self.button and self.count:
            click(buttons[self.button], self.count)
        time.sleep(0.01)
        return None

    def data(self):
        return {"MouseClick": [self.button, self.count]}


class Set(Action):
    def __init__(self, args, warn=True):
        if not (isinstance(args, list) and len(args) > 2):
            if warn:
                logger.warning("rule Set argument not list with minimum length 3: %s", args)
            self.args = []
        else:
            self.args = args

    def __str__(self):
        return "Set: " + " ".join([str(a) for a in self.args])

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if len(self.args) < 3:
            return None
        if logger.isEnabledFor(logging.INFO):
            logger.info("Set action: %s", self.args)
        dev = device.find(self.args[0]) if self.args[0] is not None else device
        if dev is None:
            logger.warning("Set action: device %s is not known", self.args[0])
            return None
        setting = next((s for s in dev.settings if s.name == self.args[1]), None)
        if setting is None:
            logger.warning("Set action: setting %s is not the name of a setting for %s", self.args[1], dev.name)
            return None
        args = setting.acceptable(self.args[2:], setting.read())
        if args is None:
            logger.warning(
                "Set Action: invalid args %s for setting %s of %s",
                self.args[2:],
                self.args[1],
                self.args[0],
            )
            return None
        if len(args) > 1:
            setting.write_key_value(args[0], args[1])
        else:
            setting.write(args[0])
        if device.setting_callback:
            device.setting_callback(device, type(setting), args)
        return None

    def data(self):
        return {"Set": self.args[:]}


class Execute(Action):
    def __init__(self, args, warn=True):
        if isinstance(args, str):
            args = [args]
        if not (isinstance(args, list) and all(isinstance(arg), str) for arg in args):
            if warn:
                logger.warning("rule Execute argument not list of strings: %s", args)
            self.args = []
        else:
            self.args = args

    def __str__(self):
        return "Execute: " + " ".join([a for a in self.args])

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if logger.isEnabledFor(logging.INFO):
            logger.info("Execute action: %s", self.args)
        subprocess.Popen(self.args)
        return None

    def data(self):
        return {"Execute": self.args[:]}


class Later(Action):
    def __init__(self, args, warn=True):
        self.delay = 0
        self.rule = Rule([])
        self.components = []
        if not (isinstance(args, list)):
            args = [args]
        if not (isinstance(args, list) and len(args) >= 1):
            if warn:
                logger.warning("rule Later argument not list with minimum length 1: %s", args)
        elif not (isinstance(args[0], (int, float))) or not 0.01 <= args[0] <= 100:
            if warn:
                logger.warning("rule Later delay not between 0.01 and 100: %s", args)
        else:
            self.delay = args[0]
            self.rule = Rule(args[1:], warn=warn)
            self.components = self.rule.components

    def __str__(self):
        return f"Later: [{str(self.delay)}, " + ", ".join(str(c) for c in self.components) + "]"

    def evaluate(self, feature, notification: HIDPPNotification, device, last_result):
        if self.delay and self.rule:
            if self.delay >= 1:
                GLib.timeout_add_seconds(int(self.delay), Rule.once, self.rule, feature, notification, device, last_result)
            else:
                GLib.timeout_add(int(self.delay * 1000), Rule.once, self.rule, feature, notification, device, last_result)
        return None

    def data(self):
        data = [c.data() for c in self.components]
        data.insert(0, self.delay)
        return {"Later": data}


COMPONENTS = {
    "Rule": Rule,
    "Not": Not,
    "Or": Or,
    "And": And,
    "Process": Process,
    "MouseProcess": MouseProcess,
    "Feature": Feature,
    "Report": Report,
    "Setting": Setting,
    "Modifiers": Modifiers,
    "Key": Key,
    "KeyIsDown": KeyIsDown,
    "Test": Test,
    "TestBytes": TestBytes,
    "MouseGesture": MouseGesture,
    "Active": Active,
    "Device": Device,
    "Host": Host,
    "KeyPress": KeyPress,
    "MouseScroll": MouseScroll,
    "MouseClick": MouseClick,
    "Set": Set,
    "Execute": Execute,
    "Later": Later,
}


built_in_rules = Rule(
    [
        {
            "Rule": [  # Implement problematic keys for Craft and MX Master
                {"Rule": [{"Key": ["Brightness Down", "pressed"]}, {"KeyPress": "XF86_MonBrightnessDown"}]},
                {"Rule": [{"Key": ["Brightness Up", "pressed"]}, {"KeyPress": "XF86_MonBrightnessUp"}]},
            ]
        },
    ]
)


def key_is_down(key: NamedInt) -> bool:
    """Checks if given key is pressed or not."""
    if key == CONTROL.MR:
        return mr_key_down
    elif CONTROL.M1 <= key <= CONTROL.M8:
        return bool(m_keys_down & (0x01 << (key - CONTROL.M1)))
    elif CONTROL.G1 <= key <= CONTROL.G32:
        return bool(g_keys_down & (0x01 << (key - CONTROL.G1)))
    return key in keys_down


def evaluate_rules(feature, notification: HIDPPNotification, device):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("evaluating rules on %s %s", feature, notification)
    rules.evaluate(feature, notification, device, True)


def process_notification(device, notification: HIDPPNotification, feature) -> None:
    """Processes HID++ notifications."""
    global keys_down, g_keys_down, m_keys_down, mr_key_down, key_down, key_up, thumb_wheel_displacement
    key_down, key_up = None, None
    # need to keep track of keys that are down to find a new key down
    if notification.address == 0x00:
        if feature == SupportedFeature.REPROG_CONTROLS_V4:
            new_keys_down = struct.unpack("!4H", notification.data[:8])
            for key in new_keys_down:
                if key and key not in keys_down:
                    key_down = key
            for key in keys_down:
                if key and key not in new_keys_down:
                    key_up = key
            keys_down = new_keys_down
        # and also G keys down
        elif feature == SupportedFeature.GKEY:
            new_g_keys_down = struct.unpack("<I", notification.data[:4])[0]
            for i in range(32):
                if new_g_keys_down & (0x01 << i) and not g_keys_down & (0x01 << i):
                    key_down = CONTROL["G" + str(i + 1)]
                if g_keys_down & (0x01 << i) and not new_g_keys_down & (0x01 << i):
                    key_up = CONTROL["G" + str(i + 1)]
            g_keys_down = new_g_keys_down
        # and also M keys down
        elif feature == SupportedFeature.MKEYS:
            new_m_keys_down = struct.unpack("!1B", notification.data[:1])[0]
            for i in range(1, 9):
                if new_m_keys_down & (0x01 << (i - 1)) and not m_keys_down & (0x01 << (i - 1)):
                    key_down = CONTROL["M" + str(i)]
                if m_keys_down & (0x01 << (i - 1)) and not new_m_keys_down & (0x01 << (i - 1)):
                    key_up = CONTROL["M" + str(i)]
            m_keys_down = new_m_keys_down
        # and also MR key
        elif feature == SupportedFeature.MR:
            new_mr_key_down = struct.unpack("!1B", notification.data[:1])[0]
            if not mr_key_down and new_mr_key_down:
                key_down = CONTROL["MR"]
            if mr_key_down and not new_mr_key_down:
                key_up = CONTROL["MR"]
            mr_key_down = new_mr_key_down
        # keep track of thumb wheel movement
        elif feature == SupportedFeature.THUMB_WHEEL:
            if notification.data[4] <= 0x01:  # when wheel starts, zero out last movement
                thumb_wheel_displacement = 0
            thumb_wheel_displacement += signed(notification.data[0:2])

    GLib.idle_add(evaluate_rules, feature, notification, device)


_XDG_CONFIG_HOME = os.environ.get("XDG_CONFIG_HOME") or os.path.expanduser(os.path.join("~", ".config"))
_file_path = os.path.join(_XDG_CONFIG_HOME, "solaar", "rules.yaml")

rules = built_in_rules


def _save_config_rule_file(file_name: str = _file_path):
    # This is a trick to show str/float/int lists in-line (inspired by https://stackoverflow.com/a/14001707)
    class inline_list(list):
        pass

    def blockseq_rep(dumper, data):
        return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)

    yaml.add_representer(inline_list, blockseq_rep)

    def convert(elem):
        if isinstance(elem, list):
            if len(elem) == 1 and isinstance(elem[0], (int, str, float)):
                # All diversion classes that expect a list of scalars also support a single scalar without a list
                return elem[0]
            if all(isinstance(c, (int, str, float)) for c in elem):
                return inline_list([convert(c) for c in elem])
            return [convert(c) for c in elem]
        if isinstance(elem, dict):
            return {k: convert(v) for k, v in elem.items()}
        if isinstance(elem, NamedInt):
            return int(elem)
        return elem

    # YAML format settings
    dump_settings = {
        "encoding": "utf-8",
        "explicit_start": True,
        "explicit_end": True,
        "default_flow_style": False,
        # 'version': (1, 3),  # it would be printed for every rule
    }
    # Save only user-defined rules
    rules_to_save = sum((r.data()["Rule"] for r in rules.components if r.source == file_name), [])
    if logger.isEnabledFor(logging.INFO):
        logger.info("saving %d rule(s) to %s", len(rules_to_save), file_name)
    try:
        with open(file_name, "w") as f:
            if rules_to_save:
                f.write("%YAML 1.3\n")  # Write version manually
            dump_data = [r["Rule"] for r in rules_to_save]
            yaml.dump_all(convert(dump_data), f, **dump_settings)
    except Exception as e:
        logger.error("failed to save to %s\n%s", file_name, e)
        return False
    return True


def load_config_rule_file():
    """Loads user configured rules."""
    global rules

    if os.path.isfile(_file_path):
        rules = _load_rule_config(_file_path)


def _load_rule_config(file_path: str) -> Rule:
    loaded_rules = []
    try:
        with open(file_path) as config_file:
            loaded_rules = []
            for loaded_rule in yaml.safe_load_all(config_file):
                rule = Rule(loaded_rule, source=file_path)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("load rule: %s", rule)
                loaded_rules.append(rule)
            if logger.isEnabledFor(logging.INFO):
                logger.info("loaded %d rules from %s", len(loaded_rules), config_file.name)
    except Exception as e:
        logger.error("failed to load from %s\n%s", file_path, e)
    return Rule([Rule(loaded_rules, source=file_path), built_in_rules])


load_config_rule_file()



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
from __future__ import annotations

import logging
import struct
import time

from enum import IntEnum
from typing import Any

from solaar.i18n import _

from . import common
from . import hidpp20_constants
from . import settings_validator
from .common import NamedInt

logger = logging.getLogger(__name__)

SENSITIVITY_IGNORE = "ignore"


class Kind(IntEnum):
    NONE = 0
    TOGGLE = 0x01
    CHOICE = 0x02
    RANGE = 0x04
    MAP_CHOICE = 0x0A
    MULTIPLE_TOGGLE = 0x10
    PACKED_RANGE = 0x20
    MULTIPLE_RANGE = 0x40
    HETERO = 0x80
    MAP_RANGE = 0x102
    COLOR = 0x200


class Setting:
    """A setting descriptor. Needs to be instantiated for each specific device."""

    name = label = description = ""
    feature = register = kind = None
    min_version = 0
    persist = True
    rw_options = {}
    validator_class = None
    validator_options = {}
    display = True  # display setting in UI

    def __init__(self, device, rw, validator):
        self._device = device
        self._rw = rw
        self._validator = validator
        self.kind = getattr(self._validator, "kind", None)
        self._value = None

    @classmethod
    def build(cls, device):
        assert cls.feature or cls.register, "Settings require either a feature or a register"
        rw_class = cls.rw_class if hasattr(cls, "rw_class") else FeatureRW if cls.feature else RegisterRW
        rw = rw_class(cls.feature if cls.feature else cls.register, **cls.rw_options)
        p = device.protocol
        if p == 1.0:  # HID++ 1.0 devices do not support features
            assert rw.kind == RegisterRW.kind
        elif p >= 2.0:  # HID++ 2.0 devices do not support registers
            assert rw.kind == FeatureRW.kind
        validator_class = cls.validator_class
        validator = validator_class.build(cls, device, **cls.validator_options)
        if validator:
            assert cls.kind is None or cls.kind & validator.kind != 0
            return cls(device, rw, validator)

    def val_to_string(self, value):
        return self._validator.to_string(value)

    @property
    def choices(self):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")

        return self._validator.choices if self._validator and self._validator.kind & Kind.CHOICE else None

    @property
    def range(self):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")

        if self._validator.kind == Kind.RANGE:
            return self._validator.min_value, self._validator.max_value

    def _pre_read(self, cached, key=None):
        if self.persist and self._value is None and getattr(self._device, "persister", None):
            # We haven't read a value from the device yet,
            # maybe we have something in the configuration.
            self._value = self._device.persister.get(self.name)
        if cached and self._value is not None:
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                # If this is a new device (or a new setting for an old device),
                # make sure to save its current value for the next time.
                self._device.persister[self.name] = self._value if self.persist else None

    def read(self, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")

        self._pre_read(cached)
        if cached and self._value is not None:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("%s: cached value %r on %s", self.name, self._value, self._device)
            return self._value

        if self._device.online:
            reply = self._rw.read(self._device)
            if reply:
                self._value = self._validator.validate_read(reply)
            if self._value is not None and self._device.persister and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value if self.persist else None
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("%s: read value %r on %s", self.name, self._value, self._device)
            return self._value

    def _pre_write(self, save=True):
        # Remember the value we're trying to set, even if the write fails.
        # This way even if the device is offline or some other error occurs,
        # the last value we've tried to write is remembered in the configuration.
        if self._device.persister and save:
            self._device.persister[self.name] = self._value if self.persist else None

    def update(self, value, save=True):
        self._value = value
        self._pre_write(save)

    def write(self, value, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert value is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: write %r to %s", self.name, value, self._device)

        if self._device.online:
            if self._value != value:
                self.update(value, save)

            current_value = None
            if self._validator.needs_current_value:
                # the _validator needs the current value, possibly to merge flag values
                current_value = self._rw.read(self._device)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: current value %r on %s", self.name, current_value, self._device)

            data_bytes = self._validator.prepare_write(value, current_value)
            if data_bytes is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: prepare write(%s) => %r", self.name, value, data_bytes)

                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    # tell whomever is calling that the write failed
                    return None

            return value

    def acceptable(self, args, current):
        return self._validator.acceptable(args, current) if self._validator else None

    def compare(self, args, current):
        return self._validator.compare(args, current) if self._validator else None

    def apply(self):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: apply (%s)", self.name, self._device)
        try:
            value = self.read(self.persist)  # Don't use persisted value if setting doesn't persist
            if self.persist and value is not None:  # If setting doesn't persist no need to write value just read
                self.write(value, save=False)
        except Exception as e:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning("%s: error applying %s so ignore it (%s): %s", self.name, self._value, self._device, repr(e))

    def __str__(self):
        if hasattr(self, "_value"):
            assert hasattr(self, "_device")
            return "<Setting([%s:%s] %s:%s=%s)>" % (
                self._rw.kind,
                self._validator.kind if self._validator else None,
                self._device.codename,
                self.name,
                self._value,
            )
        return f"<Setting([{self._rw.kind}:{self._validator.kind if self._validator else None}] {self.name})>"

    __repr__ = __str__


class Settings(Setting):
    """A setting descriptor for multiple choices, being a map from keys to values.
    Needs to be instantiated for each specific device."""

    def read(self, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r from %s", self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply_map = {}
            for key in self._validator.choices:
                reply = self._rw.read(self._device, key)
                if reply:
                    reply_map[int(key)] = self._validator.validate_read(reply, key)
            self._value = reply_map
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value

    def read_key(self, key, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert key is not None
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r key %r from %s", self.name, self._value, key, self._device)

        self._pre_read(cached)
        if cached and self._value is not None:
            return self._value[int(key)]

        if self._device.online:
            reply = self._rw.read(self._device, key)
            if reply:
                self._value[int(key)] = self._validator.validate_read(reply, key)
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value[int(key)]

    def write(self, map, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert map is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings write %r to %s", self.name, map, self._device)

        if self._device.online:
            self.update(map, save)
            for key, value in map.items():
                data_bytes = self._validator.prepare_write(int(key), value)
                if data_bytes is not None:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("%s: settings prepare map write(%s,%s) => %r", self.name, key, value, data_bytes)
                    reply = self._rw.write(self._device, int(key), data_bytes)
                    if not reply:
                        return None
            return map

    def update_key_value(self, key, value, save=True):
        self._value[int(key)] = value
        self._pre_write(save)

    def write_key_value(self, key, value, save=True) -> Any | None:
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert key is not None
        assert value is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings write key %r value %r to %s", self.name, key, value, self._device)

        if self._device.online:
            if not self._value:
                self.read()
            try:
                data_bytes = self._validator.prepare_write(int(key), value)
                # always need to write to configuration because dictionary is shared and could have changed
                self.update_key_value(key, value, save)
            except ValueError:
                data_bytes = value = None
            if data_bytes is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: settings prepare key value write(%s,%s) => %r", self.name, key, value, data_bytes)
                reply = self._rw.write(self._device, int(key), data_bytes)
                if not reply:
                    return None
            return value


class LongSettings(Setting):
    """A setting descriptor for multiple choices, being a map from keys to values.
    Allows multiple write requests, if the options don't fit in 16 bytes.
    The validator must return a list.
    Needs to be instantiated for each specific device."""

    def read(self, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r from %s", self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply_map = {}
            # Reading one item at a time. This can probably be optimised
            for item in self._validator.items:
                r = self._validator.prepare_read_item(item)
                reply = self._rw.read(self._device, r)
                if reply:
                    reply_map[int(item)] = self._validator.validate_read_item(reply, item)
            self._value = reply_map
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value

    def read_item(self, item, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert item is not None
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r item %r from %s", self.name, self._value, item, self._device)

        self._pre_read(cached)
        if cached and self._value is not None:
            return self._value[int(item)]

        if self._device.online:
            r = self._validator.prepare_read_item(item)
            reply = self._rw.read(self._device, r)
            if reply:
                self._value[int(item)] = self._validator.validate_read_item(reply, item)
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value[int(item)]

    def write(self, map, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert map is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: long settings write %r to %s", self.name, map, self._device)
        if self._device.online:
            self.update(map, save)
            for item, value in map.items():
                data_bytes_list = self._validator.prepare_write(self._value)
                if data_bytes_list is not None:
                    for data_bytes in data_bytes_list:
                        if data_bytes is not None:
                            if logger.isEnabledFor(logging.DEBUG):
                                logger.debug("%s: settings prepare map write(%s,%s) => %r", self.name, item, value, data_bytes)
                            reply = self._rw.write(self._device, data_bytes)
                            if not reply:
                                return None
            return map

    def update_key_value(self, key, value, save=True):
        self._value[int(key)] = value
        self._pre_write(save)

    def write_key_value(self, item, value, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert item is not None
        assert value is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: long settings write item %r value %r to %s", self.name, item, value, self._device)

        if self._device.online:
            if not self._value:
                self.read()
            data_bytes = self._validator.prepare_write_item(item, value)
            self.update_key_value(item, value, save)
            if data_bytes is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: settings prepare item value write(%s,%s) => %r", self.name, item, value, data_bytes)
                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    return None
            return value


class BitFieldSetting(Setting):
    """A setting descriptor for a set of choices represented by one bit each, being a map from options to booleans.
    Needs to be instantiated for each specific device."""

    def read(self, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r from %s", self.name, self._value, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value

        if self._device.online:
            reply_map = {}
            reply = self._do_read()
            if reply:
                reply_map = self._validator.validate_read(reply)
            self._value = reply_map
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value

    def _do_read(self):
        return self._rw.read(self._device)

    def read_key(self, key, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert key is not None
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r key %r from %s", self.name, self._value, key, self._device)

        self._pre_read(cached)

        if cached and self._value is not None:
            return self._value[int(key)]

        if self._device.online:
            reply = self._do_read_key(key)
            if reply:
                self._value = self._validator.validate_read(reply)
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value[int(key)]

    def _do_read_key(self, key):
        return self._rw.read(self._device, key)

    def write(self, map, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert map is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: bit field settings write %r to %s", self.name, map, self._device)
        if self._device.online:
            self.update(map, save)
            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: settings prepare map write(%s) => %r", self.name, self._value, data_bytes)
                # if prepare_write returns a list, write one item at a time
                seq = data_bytes if isinstance(data_bytes, list) else [data_bytes]
                for b in seq:
                    reply = self._rw.write(self._device, b)
                    if not reply:
                        return None
            return map

    def update_key_value(self, key, value, save=True):
        self._value[int(key)] = value
        self._pre_write(save)

    def write_key_value(self, key, value, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert key is not None
        assert value is not None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: bit field settings write key %r value %r to %s", self.name, key, value, self._device)

        if self._device.online:
            if not self._value:
                self.read()
            value = bool(value)
            self.update_key_value(key, value, save)

            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: settings prepare key value write(%s,%s) => %r", self.name, key, str(value), data_bytes)
                # if prepare_write returns a list, write one item at a time
                seq = data_bytes if isinstance(data_bytes, list) else [data_bytes]
                for b in seq:
                    reply = self._rw.write(self._device, b)
                    if not reply:
                        return None

            return value


class BitFieldWithOffsetAndMaskSetting(BitFieldSetting):
    """A setting descriptor for a set of choices represented by one bit each,
    each one having an offset, being a map from options to booleans.
    Needs to be instantiated for each specific device."""

    def _do_read(self):
        return {r: self._rw.read(self._device, r) for r in self._validator.prepare_read()}

    def _do_read_key(self, key):
        r = self._validator.prepare_read_key(key)
        return {r: self._rw.read(self._device, r)}


class RangeFieldSetting(Setting):
    """A setting descriptor for a set of choices represented by one field each, with map from option names to range(0,n).
    Needs to be instantiated for each specific device."""

    def read(self, cached=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: settings read %r from %s", self.name, self._value, self._device)
        self._pre_read(cached)
        if cached and self._value is not None:
            return self._value
        if self._device.online:
            reply_map = {}
            reply = self._do_read()
            if reply:
                reply_map = self._validator.validate_read(reply)
            self._value = reply_map
            if getattr(self._device, "persister", None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value if self.persist else None
            return self._value

    def _do_read(self):
        return self._rw.read(self._device)

    def read_key(self, key, cached=True):
        return self.read(cached)[int(key)]

    def write(self, map, save=True):
        assert hasattr(self, "_value")
        assert hasattr(self, "_device")
        assert map is not None
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: range field setting write %r to %s", self.name, map, self._device)
        if self._device.online:
            self.update(map, save)
            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: range field setting prepare map write(%s) => %r", self.name, self._value, data_bytes)
                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    return None
            elif logger.isEnabledFor(logging.WARNING):
                logger.warning("%s: range field setting no data to write", self.name)
            return map

    def write_key_value(self, key, value, save=True):
        assert key is not None
        assert value is not None
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: range field setting write key %r value %r to %s", self.name, key, value, self._device)
        if self._device.online:
            if not self._value:
                self.read()
            map = self._value
            map[int(key)] = value
            self.write(map, save)
            return value


#
# read/write low-level operators
#


class RegisterRW:
    __slots__ = ("register",)

    kind = NamedInt(0x01, _("register"))

    def __init__(self, register: int):
        assert isinstance(register, int)
        self.register = register

    def read(self, device):
        return device.read_register(self.register)

    def write(self, device, data_bytes):
        return device.write_register(self.register, data_bytes)


class FeatureRW:
    kind = NamedInt(0x02, _("feature"))
    default_read_fnid = 0x00
    default_write_fnid = 0x10

    def __init__(
        self,
        feature: hidpp20_constants.SupportedFeature,
        read_fnid=0x00,
        write_fnid=0x10,
        prefix=b"",
        suffix=b"",
        read_prefix=b"",
        no_reply=False,
    ):
        assert isinstance(feature, hidpp20_constants.SupportedFeature)
        self.feature = feature
        self.read_fnid = read_fnid
        self.write_fnid = write_fnid
        self.no_reply = no_reply
        self.prefix = prefix
        self.suffix = suffix
        self.read_prefix = read_prefix

    def read(self, device, data_bytes=b""):
        assert self.feature is not None
        if self.read_fnid is not None:
            return device.feature_request(self.feature, self.read_fnid, self.prefix, self.read_prefix, data_bytes)
        else:
            return b""

    def write(self, device, data_bytes):
        assert self.feature is not None
        write_bytes = self.prefix + (data_bytes.to_bytes(1) if isinstance(data_bytes, int) else data_bytes) + self.suffix
        reply = device.feature_request(self.feature, self.write_fnid, write_bytes, no_reply=self.no_reply)
        return reply if not self.no_reply else True


class FeatureRWMap(FeatureRW):
    kind = NamedInt(0x02, _("feature"))
    default_read_fnid = 0x00
    default_write_fnid = 0x10
    default_key_byte_count = 1

    def __init__(
        self,
        feature: hidpp20_constants.SupportedFeature,
        read_fnid=default_read_fnid,
        write_fnid=default_write_fnid,
        key_byte_count=default_key_byte_count,
        no_reply=False,
    ):
        assert isinstance(feature, hidpp20_constants.SupportedFeature)
        self.feature = feature
        self.read_fnid = read_fnid
        self.write_fnid = write_fnid
        self.key_byte_count = key_byte_count
        self.no_reply = no_reply

    def read(self, device, key):
        assert self.feature is not None
        key_bytes = common.int2bytes(key, self.key_byte_count)
        return device.feature_request(self.feature, self.read_fnid, key_bytes)

    def write(self, device, key, data_bytes):
        assert self.feature is not None
        key_bytes = common.int2bytes(key, self.key_byte_count)
        reply = device.feature_request(self.feature, self.write_fnid, key_bytes, data_bytes, no_reply=self.no_reply)
        return reply if not self.no_reply else True


class ActionSettingRW:
    """Special RW class for settings that turn on and off special processing when a key or button is depressed"""

    def __init__(self, feature, name="", divert_setting_name="divert-keys"):
        self.feature = feature  # not used?
        self.name = name
        self.divert_setting_name = divert_setting_name
        self.kind = FeatureRW.kind  # pretend to be FeatureRW as required for HID++ 2.0 devices
        self.device = None
        self.key = None
        self.active = False
        self.pressed = False

    def activate_action(self):  # action to take when setting is activated (write non-false)
        pass

    def deactivate_action(self):  # action to take when setting is deactivated (write false)
        pass

    def press_action(self):  # action to take when key is pressed
        pass

    def release_action(self):  # action to take when key is released
        pass

    def move_action(self, dx, dy):  # action to take when mouse is moved while key is down
        pass

    def key_action(self, key):  # acction to take when some other diverted key is pressed
        pass

    def read(self, device):  # need to return bytes, as if read from device
        return common.int2bytes(self.key.key, 2) if self.active and self.key else b"\x00\x00"

    def write(self, device, data_bytes):
        def handler(device, n):  # Called on notification events from the device
            if (
                n.sub_id < 0x40
                and device.features.get_feature(n.sub_id) == hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4
            ):
                if n.address == 0x00:
                    cids = struct.unpack("!HHHH", n.data[:8])
                    if not self.pressed and int(self.key.key) in cids:  # trigger key pressed
                        self.pressed = True
                        self.press_action()
                    elif self.pressed:
                        if int(self.key.key) not in cids:  # trigger key released
                            self.pressed = False
                            self.release_action()
                        else:
                            for key in cids:
                                if key and not key == self.key.key:  # some other diverted key pressed
                                    self.key_action(key)
                elif n.address == 0x10:
                    if self.pressed:
                        dx, dy = struct.unpack("!hh", n.data[:4])
                        self.move_action(dx, dy)

        divertSetting = next(filter(lambda s: s.name == self.divert_setting_name, device.settings), None)
        if divertSetting is None:
            logger.warning("setting %s not found on %s", self.divert_setting_name, device.name)
            return None
        self.device = device
        key = common.bytes2int(data_bytes)
        if key:  # Enable
            self.key = next((k for k in device.keys if k.key == key), None)
            if self.key:
                self.active = True
                if divertSetting:
                    divertSetting.write_key_value(int(self.key.key), 1)
                    if self.device.setting_callback:
                        self.device.setting_callback(device, type(divertSetting), [self.key.key, 1])
                device.add_notification_handler(self.name, handler)
                self.activate_action()
            else:
                logger.error("cannot enable %s on %s for key %s", self.name, device, key)
        else:  # Disable
            if self.active:
                self.active = False
                if divertSetting:
                    divertSetting.write_key_value(int(self.key.key), 0)
                    if self.device.setting_callback:
                        self.device.setting_callback(device, type(divertSetting), [self.key.key, 0])
                try:
                    device.remove_notification_handler(self.name)
                except Exception:
                    if logger.isEnabledFor(logging.WARNING):
                        logger.warning("cannot disable %s on %s", self.name, device)
                self.deactivate_action()
        return data_bytes


class RawXYProcessing:
    """Special class for processing RawXY action messages initiated by pressing a key with rawXY diversion capability"""

    def __init__(self, device, name=""):
        self.device = device
        self.name = name
        self.keys = []  # the keys that can initiate processing
        self.initiating_key = None  # the key that did initiate processing
        self.active = False
        self.feature_offset = device.features[hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4]
        assert self.feature_offset is not False

    def handler(self, device, n):  # Called on notification events from the device
        if n.sub_id < 0x40 and device.features.get_feature(n.sub_id) == hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4:
            if n.address == 0x00:
                cids = struct.unpack("!HHHH", n.data[:8])
                ## generalize to list of keys
                if not self.initiating_key:  # no initiating key pressed
                    for k in self.keys:
                        if int(k.key) in cids:  # initiating key that was pressed
                            self.initiating_key = k
                    if self.initiating_key:
                        self.press_action(self.initiating_key)
                else:
                    if int(self.initiating_key.key) not in cids:  # initiating key released
                        self.initiating_key = None
                        self.release_action()
                    else:
                        for key in cids:
                            if key and key != self.initiating_key.key:
                                self.key_action(key)
            elif n.address == 0x10:
                if self.initiating_key:
                    dx, dy = struct.unpack("!hh", n.data[:4])
                    self.move_action(dx, dy)

    def start(self, key):
        device_key = next((k for k in self.device.keys if k.key == key), None)
        if device_key:
            self.keys.append(device_key)
            if not self.active:
                self.active = True
                self.activate_action()
                self.device.add_notification_handler(self.name, self.handler)
            device_key.set_rawXY_reporting(True)

    def stop(self, key):  # only stop if this is the active key
        if self.active:
            processing_key = next((k for k in self.keys if k.key == key), None)
            if processing_key:
                processing_key.set_rawXY_reporting(False)
                self.keys.remove(processing_key)
            if not self.keys:
                try:
                    self.device.remove_notification_handler(self.name)
                except Exception:
                    if logger.isEnabledFor(logging.WARNING):
                        logger.warning("cannot disable %s on %s", self.name, self.device)
                self.deactivate_action()
                self.active = False

    def activate_action(self):  # action to take when processing is activated
        pass

    def deactivate_action(self):  # action to take when processing is deactivated
        pass

    def press_action(self, key):  # action to take when an initiating key is pressed
        pass

    def release_action(self):  # action to take when key is released
        pass

    def move_action(self, dx, dy):  # action to take when mouse is moved while key is down
        pass

    def key_action(self, key):  # acction to take when some other diverted key is pressed
        pass


def apply_all_settings(device):
    if device.features and hidpp20_constants.SupportedFeature.HIRES_WHEEL in device.features:
        time.sleep(0.2)  # delay to try to get out of race condition with Linux HID++ driver
    persister = getattr(device, "persister", None)
    sensitives = persister.get("_sensitive", {}) if persister else {}
    for s in device.settings:
        ignore = sensitives.get(s.name, False)
        if ignore != SENSITIVITY_IGNORE:
            s.apply()


Setting.validator_class = settings_validator.BooleanValidator



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

from enum import Enum
from typing import Callable

import gi
import yaml

from logitech_receiver.common import Alert

from solaar.i18n import _
from solaar.ui.config_panel import change_setting
from solaar.ui.config_panel import record_setting
from solaar.ui.window import find_device

from . import common
from . import desktop_notifications
from . import diversion_rules
from . import tray
from . import window

gi.require_version("Gtk", "3.0")
from gi.repository import Gio  # NOQA: E402
from gi.repository import GLib  # NOQA: E402
from gi.repository import Gtk  # NOQA: E402

logger = logging.getLogger(__name__)

assert Gtk.get_major_version() > 2, "Solaar requires Gtk 3 python bindings"


APP_ID = "io.github.pwr_solaar.solaar"


class GtkSignal(Enum):
    ACTIVATE = "activate"
    COMMAND_LINE = "command-line"
    SHUTDOWN = "shutdown"


def _startup(app, startup_hook, use_tray, show_window):
    logger.debug("startup registered=%s, remote=%s", app.get_is_registered(), app.get_is_remote())
    common.start_async()
    desktop_notifications.init()
    if use_tray:
        tray.init(lambda _ignore: window.destroy())
    window.init(show_window, use_tray)
    startup_hook()


def _activate(app):
    logger.debug("activate")
    if app.get_windows():
        window.popup()
    else:
        app.add_window(window._window)


def _command_line(app, command_line):
    args = command_line.get_arguments()
    args = yaml.safe_load("".join(args)) if args else args
    if not args:
        _activate(app)
    elif args[0] == "config":  # config call from remote instance
        logger.info("remote command line %s", args)
        dev = find_device(args[1])
        if dev:
            setting = next((s for s in dev.settings if s.name == args[2]), None)
            if setting:
                change_setting(dev, setting, args[3:])
    return 0


def _shutdown(_app, shutdown_hook):
    logger.debug("shutdown")
    shutdown_hook()
    common.stop_async()
    tray.destroy()
    desktop_notifications.uninit()


def run_loop(
    startup_hook: Callable[[], None],
    shutdown_hook: Callable[[], None],
    use_tray: bool,
    show_window: bool,
):
    assert use_tray or show_window, "need either tray or visible window"

    application = Gtk.Application.new(APP_ID, Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

    application.connect(
        "startup",
        lambda app, startup_hook: _startup(app, startup_hook, use_tray, show_window),
        startup_hook,
    )
    application.connect(GtkSignal.COMMAND_LINE.value, _command_line)
    application.connect(GtkSignal.ACTIVATE.value, _activate)
    application.connect(GtkSignal.SHUTDOWN.value, _shutdown, shutdown_hook)

    application.register()
    if application.get_is_remote():
        print(_("Another Solaar process is already running so just expose its window"))
    application.run()


def _status_changed(device, alert, reason, refresh=False):
    assert device is not None
    logger.debug("status changed: %s (%s) %s", device, alert, reason)
    if alert is None:
        alert = Alert.NONE

    tray.update(device)
    if alert & Alert.ATTENTION:
        tray.attention(reason)

    need_popup = alert & Alert.SHOW_WINDOW
    window.update(device, need_popup, refresh)
    diversion_rules.update_devices()

    if alert & (Alert.NOTIFICATION | Alert.ATTENTION):
        desktop_notifications.show(device, reason)


def status_changed(device, alert=Alert.NONE, reason=None, refresh=False):
    GLib.idle_add(_status_changed, device, alert, reason, refresh)


def setting_changed(device, setting_class, vals):
    record_setting(device, setting_class, vals)



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import yaml

from logitech_receiver import settings
from logitech_receiver import settings_templates
from logitech_receiver.common import NamedInts
from logitech_receiver.settings_templates import SettingsProtocol

from solaar import configuration

APP_ID = "io.github.pwr_solaar.solaar"


def _print_setting(s, verbose=True):
    print("#", s.label)
    if verbose:
        if s.description:
            print("#", s.description.replace("\n", " "))
        if s.kind == settings.Kind.TOGGLE:
            print("#   possible values: on/true/t/yes/y/1 or off/false/f/no/n/0 or Toggle/~")
        elif s.kind == settings.Kind.CHOICE:
            print(
                "#   possible values: one of [",
                ", ".join(str(v) for v in s.choices),
                "], or higher/lower/highest/max/lowest/min",
            )
        else:
            # wtf?
            pass
    value = s.read(cached=False)
    if value is None:
        print(s.name, "= ? (failed to read from device)")
    else:
        print(s.name, "=", s.val_to_string(value))


def _print_setting_keyed(s, key, verbose=True):
    print("#", s.label)
    if verbose:
        if s.description:
            print("#", s.description.replace("\n", " "))
        if s.kind == settings.Kind.MULTIPLE_TOGGLE:
            k = next((k for k in s._labels if key == k), None)
            if k is None:
                print(s.name, "=? (key not found)")
            else:
                print("#   possible values: on/true/t/yes/y/1 or off/false/f/no/n/0 or Toggle/~")
                value = s.read(cached=False)
                if value is None:
                    print(s.name, "= ? (failed to read from device)")
                else:
                    print(s.name, s.val_to_string({k: value[str(int(k))]}))
        elif s.kind == settings.Kind.MAP_CHOICE:
            k = next((k for k in s.choices.keys() if key == k), None)
            if k is None:
                print(s.name, "=? (key not found)")
            else:
                print("#   possible values: one of [", ", ".join(str(v) for v in s.choices[k]), "]")
                value = s.read(cached=False)
                if value is None:
                    print(s.name, "= ? (failed to read from device)")
                else:
                    print(s.name, s.val_to_string({k: value[int(k)]}))


def to_int(s):
    try:
        return int(s)
    except ValueError:
        return None


def select_choice(value, choices, setting, key):
    lvalue = value.lower()
    ivalue = to_int(value)
    val = None
    for choice in choices:
        if value == str(choice):
            val = choice
            break
    if val is not None:
        value = val
    elif ivalue is not None and 1 <= ivalue <= len(choices):
        value = choices[ivalue - 1]
    elif lvalue in ("higher", "lower"):
        old_value = setting.read() if key is None else setting.read_key(key)
        if old_value is None:
            raise Exception(f"{setting.name}: could not read current value")
        if lvalue == "lower":
            lower_values = choices[:old_value]
            value = lower_values[-1] if lower_values else choices[:][0]
        elif lvalue == "higher":
            higher_values = choices[old_value + 1 :]
            value = higher_values[0] if higher_values else choices[:][-1]
    elif lvalue in ("highest", "max", "first"):
        value = choices[:][-1]
    elif lvalue in ("lowest", "min", "last"):
        value = choices[:][0]
    else:
        raise Exception(f"{setting.name}: possible values are [{', '.join(str(v) for v in choices)}]")
    return value


def select_toggle(value, setting, key=None):
    if value.lower() in ("toggle", "~"):
        value = not (setting.read() if key is None else setting.read()[str(int(key))])
    else:
        try:
            value = bool(int(value))
        except Exception as exc:
            if value.lower() in ("true", "yes", "on", "t", "y"):
                value = True
            elif value.lower() in ("false", "no", "off", "f", "n"):
                value = False
            else:
                raise Exception(f"{setting.name}: don't know how to interpret '{value}' as boolean") from exc
    return value


def select_range(value, setting):
    try:
        value = int(value)
    except ValueError as exc:
        raise Exception(f"{setting.name}: can't interpret '{value}' as integer") from exc
    minimum, maximum = setting.range
    if value < minimum or value > maximum:
        raise Exception(f"{setting.name}: value '{value}' out of bounds")
    return value


def run(receivers, args, _find_receiver, find_device):
    assert receivers
    assert args.device

    device_name = args.device.lower()

    dev = None
    for dev in find_device(receivers, device_name):
        if dev.ping():
            break
        dev = None

    if not dev:
        raise Exception(f"no online device found matching '{device_name}'")

    if not args.setting:  # print all settings, so first set them all up
        if not dev.settings:
            raise Exception(f"no settings for {dev.name}")
        configuration.attach_to(dev)
        print(dev.name, f"({dev.codename}) [{dev.wpid}:{dev.serial}]")
        for s in dev.settings:
            print("")
            _print_setting(s)
        return

    setting_name = args.setting.lower()
    setting = settings_templates.check_feature_setting(dev, setting_name)
    if not setting and dev.descriptor and dev.descriptor.settings:
        for sclass in dev.descriptor.settings:
            if sclass.register and sclass.name == setting_name:
                try:
                    setting = sclass.build(dev)
                except Exception:
                    setting = None
    if setting is None:
        raise Exception(f"no setting '{args.setting}' for {dev.name}")

    if args.value_key is None:
        _print_setting(setting)
        return

    remote = False
    try:
        import gi

        gi.require_version("Gtk", "3.0")
        from gi.repository import Gio
        from gi.repository import Gtk

        if Gtk.init_check()[0]:  # can Gtk be initialized?
            application = Gtk.Application.new(APP_ID, Gio.ApplicationFlags.HANDLES_COMMAND_LINE)
            application.register()
            remote = application.get_is_remote()
    except Exception:
        pass

    # don't save configuration here because we might want the Solaar GUI to make the change
    result, message, value = set(dev, setting, args, False)
    if message is not None:
        print(message)
        if result is None:
            raise Exception(f"{setting.name}: failed to set value '{str(value)}' [{value!r}]")

    # if the Solaar UI is running tell it to also perform the set, otherwise save the change in the configuration file
    if remote:
        argl = ["config", dev.serial or dev.unitId, setting.name]
        argl.extend([a for a in [args.value_key, args.extra_subkey, args.extra2] if a is not None])
        args = yaml.dump(argl)
        application.run(args)
    else:
        if dev.persister and setting.persist:
            dev.persister[setting.name] = setting._value


def set(dev, setting: SettingsProtocol, args, save):
    if setting.kind == settings.Kind.TOGGLE:
        value = select_toggle(args.value_key, setting)
        args.value_key = value
        message = f"Setting {setting.name} of {dev.name} to {value}"
        result = setting.write(value, save=save)

    elif setting.kind == settings.Kind.RANGE:
        value = select_range(args.value_key, setting)
        args.value_key = value
        message = f"Setting {setting.name} of {dev.name} to {value}"
        result = setting.write(value, save=save)

    elif setting.kind == settings.Kind.CHOICE:
        value = select_choice(args.value_key, setting.choices, setting, None)
        args.value_key = int(value)
        message = f"Setting {setting.name} of {dev.name} to {value}"
        result = setting.write(value, save=save)

    elif setting.kind == settings.Kind.MAP_CHOICE:
        if args.extra_subkey is None:
            _print_setting_keyed(setting, args.value_key)
            return None, None, None
        key = args.value_key
        ikey = to_int(key)
        k = next((k for k in setting.choices.keys() if key == k), None)
        if k is None and ikey is not None:
            k = next((k for k in setting.choices.keys() if ikey == k), None)
        if k is not None:
            value = select_choice(args.extra_subkey, setting.choices[k], setting, key)
            args.extra_subkey = int(value)
            args.value_key = str(int(k))
        else:
            raise Exception(f"{setting.name}: key '{key}' not in setting")
        message = f"Setting {setting.name} of {dev.name} key {k!r} to {value!r}"
        result = setting.write_key_value(int(k), value, save=save)

    elif setting.kind == settings.Kind.MULTIPLE_TOGGLE:
        if args.extra_subkey is None:
            _print_setting_keyed(setting, args.value_key)
            return None, None, None
        key = args.value_key
        all_keys = getattr(setting, "choices_universe", None)
        ikey = all_keys[int(key) if key.isdigit() else key] if isinstance(all_keys, NamedInts) else to_int(key)
        k = next((k for k in setting._labels if key == k), None)
        if k is None and ikey is not None:
            k = next((k for k in setting._labels if ikey == k), None)
        if k is not None:
            value = select_toggle(args.extra_subkey, setting, key=k)
            args.extra_subkey = value
            args.value_key = str(int(k))
        else:
            raise Exception(f"{setting.name}: key '{key}' not in setting")
        message = f"Setting {setting.name} key {k!r} to {value!r}"
        result = setting.write_key_value(str(int(k)), value, save=save)

    elif setting.kind == settings.Kind.MULTIPLE_RANGE:
        if args.extra_subkey is None:
            raise Exception(f"{setting.name}: setting needs both key and value to set")
        key = args.value_key
        all_keys = getattr(setting, "choices_universe", None)
        ikey = all_keys[int(key) if key.isdigit() else key] if isinstance(all_keys, NamedInts) else to_int(key)
        if args.extra2 is None or to_int(args.extra2) is None:
            raise Exception(f"{setting.name}: setting needs an integer value, not {args.extra2}")
        if not setting._value:  # ensure that there are values to look through
            setting.read()
        k = next((k for k in setting._value if key == ikey or key.isdigit() and ikey == int(key)), None)
        if k is None and ikey is not None:
            k = next((k for k in setting._value if ikey == k), None)
        item = setting._value[k]
        if args.extra_subkey in item.keys():
            item[args.extra_subkey] = to_int(args.extra2)
            args.value_key = str(int(k))
        else:
            raise Exception(f"{setting.name}: key '{key}' not in setting")
        message = f"Setting {setting.name} key {k} parameter {args.extra_subkey} to {item[args.extra_subkey]!r}"
        result = setting.write_key_value(int(k), item, save=save)
        value = item

    elif setting.kind == settings.Kind.MAP_RANGE:
        if args.extra_subkey is None:
            _print_setting_keyed(setting, args.value_key)
            return None, None, None
        key = int(args.value_key)
        value = int(args.extra_subkey)
        if key not in setting._device_object:
            raise Exception(f"{setting.name}: key '{key}' not in setting")
        message = f"Setting {setting.name} of {dev.name} key {key} to {value}"
        result = setting.write_key_value(key, value, save=save)

    elif setting.kind == settings.Kind.HETERO:
        value = yaml.safe_load(args.value_key)
        args.value_key = value
        message = f"Setting {setting.name} of {dev.name} to {value}"
        result = setting.write(value, save=save)

    else:
        print(f"Setting {setting.name}, with kind {setting.kind.name}, not implemented")
        raise Exception("NotImplemented")

    return result, message, value
